/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK_dims.h                                               */
/*  Abstract     : I/O toolkit header file                                   */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
#ifndef PRIVATE_DIMS_FASTER_H
#define PRIVATE_DIMS_FASTER_H

#include <stdint.h>
#include <algorithm>
#include <vector>
#include <deque>
#include "array.hpp"


namespace Array {


namespace Faster {

/// int values[3 * 4] = {};
/// calc_index(NDims<2>(3, 4), NDims<3>(1, 2));
/// => values[1][2] => values[1 * 4 + 2]
/// 
/// 
/// 
/// @endcode
template<class Dims>
std::size_t calc_index(const Dims &dims, const Dims &indexes){
    std::size_t index = 0;
    for(std::size_t i = 0; i < dims.size(); i++){
        std::size_t indexbuf = indexes[i];
        for(std::size_t j = i + 1; j < dims.size(); j++){
            if (dims[j]) {
                indexbuf *= dims[j];
            } else {
                break;
            }
        }
        index += indexbuf;
    }
    return index;
}

template<class Dims>
std::size_t calc_index_line(const Dims &dims, const Dims &indexes){
    if (dims.empty()) return 0;
    Dims line_dims;
    line_dims.assign(dims.begin(), dims.end() - 1);
    return calc_index(line_dims, indexes);
}

/// Set value of N-dimension.
template<class Dims, class Container, class Value>
void set_value(
    const Dims &dims, const Dims &indexes, Container &values, Value value) {
    values[calc_index(dims, indexes)] = value;
}

/// Set value of N-dimension by iterator.
template<class Dims, class InputIterator, class Value>
void set_value_it(
    const Dims &dims, const Dims &indexes, InputIterator &values_it, Value value) {
    *(values_it + calc_index(dims, indexes)) = value;
}

/// Get value of N-dimension.
template<class Dims, class Container>
typename Container::value_type get_value(
    const Dims &dims, const Dims &indexes, Container &values) {
    return values[calc_index(dims, indexes)];
}

/// Get value of N-dimension.
template<class Dims, class Container>
typename Container::value_type get_value(
    const Dims &dims, const Dims &indexes, const Container &values) {
    return values[calc_index(dims, indexes)];
}

/// Get value from iterator.
template<class Dims, class InputIterator>
typename InputIterator::value_type get_value_it(
    const Dims &dims, const Dims &indexes, InputIterator values_it) {
    return *(values_it + calc_index(dims, indexes));
}

/// Get value iterator.
template<class Dims, class InputIterator>
InputIterator get_values(
    const Dims &dims, const Dims &indexes, InputIterator value_it) {
    std::size_t index = calc_index(dims, indexes);
    return value_it + index;
}

/// Get value iterator by container.
template<class Dims, class T>
typename std::vector<T>::iterator get_values(
    const Dims &dims, const Dims &indexes, std::vector<T> &values) {
    return get_values(dims, indexes, values.begin());
}

/// Get value iterator by container.
template<class Dims, class T>
typename std::vector<T>::const_iterator get_values(
    const Dims &dims, const Dims &indexes, const std::vector<T> &values) {
    return get_values(dims, indexes, values.begin());
}

/// Get value iterator by container.
template<class Dims, class T>
typename std::deque<T>::iterator get_values(
    const Dims &dims, const Dims &indexes, std::deque<T> &values) {
    return get_values(dims, indexes, values.begin());
}

/// Get value iterator by container.
template<class Dims, class T>
typename std::deque<T>::const_iterator get_values(
    const Dims &dims, const Dims &indexes, const std::deque<T> &values) {
    return get_values(dims, indexes, values.begin());
}


/// Make boost::array<T, 1>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1) {
    boost::array<T, N> r = {{v1}}; return r;
}
/// Make boost::array<T, 2>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2) {
    boost::array<T, N> r = {{v1, v2}}; return r;
}
/// Make boost::array<T, 3>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2, const T &v3) {
    boost::array<T, N> r = {{v1, v2, v3}}; return r;
}
/// Make boost::array<T, 4>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2, const T &v3, const T &v4) {
    boost::array<T, N> r = {{v1, v2, v3, v4}}; return r;
}
/// Make boost::array<T, 5>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2, const T &v3, const T &v4, const T &v5) {
    boost::array<T, N> r = {{v1, v2, v3, v4, v5}}; return r;
}
/// Make boost::array<T, 6>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2, const T &v3, const T &v4, const T &v5, const T &v6) {
    boost::array<T, N> r = {{v1, v2, v3, v4, v5, v6}}; return r;
}
/// Make boost::array<T, 7>.
template<class T, std::size_t N> boost::array<T, N> make_array(const T &v1, const T &v2, const T &v3, const T &v4, const T &v5, const T &v6, const T &v7) {
    boost::array<T, N> r = {{v1, v2, v3, v4, v5, v6, v7}}; return r;
}

/**
   @code
   std::vector<int> values(3 * 3 * 3) = {...};
   int begin = get_value(Dims(3, 3, 3), Dims(0, 0, 0), values); // => int begin = values[0][0][0];
   int end   = get_value(Dims(3, 3, 3), Dims(2, 2, 2), values); // => int end   = values[2][2][2];
   set_value(Dims(3, 3, 3), Dims(0, 0, 0), values, 123456);     // => values[0][0][0] = 123456;
   @endcode
 *
 */
class Dims {
public:
    typedef std::size_t S;

    Dims() : dims(), size_n(0) {}

    Dims(S a)
      : dims(make_array<S, 7>(a)), size_n(1) {}
    Dims(S a, S b) 
      : dims(make_array<S, 7>(a, b)), size_n(2) {}
    Dims(S a, S b, S c)
      : dims(make_array<S, 7>(a, b, c)), size_n(3) {}
    Dims(S a, S b, S c, S d)
      : dims(make_array<S, 7>(a, b, c, d)), size_n(4) {}
    Dims(S a, S b, S c, S d, S e)
      : dims(make_array<S, 7>(a, b, c, d, e)), size_n(5) {}
    Dims(S a, S b, S c, S d, S e, S f)
      : dims(make_array<S, 7>(a, b, c, d, e, f)), size_n(6) {}
    Dims(S a, S b, S c, S d, S e, S f, S g)
      : dims(make_array<S, 7>(a, b, c, d, e, f, g)), size_n(7) {}

    template <class Container>
    void assign(Container &values) {
        for (size_t i = 0; i < values.size(); ++i) {
            if (i >= dims.size()) break;
            dims[i] = static_cast<S>(values[i]);
            ++size_n;
        }
    }

    template <class InputIterator>
    void assign(InputIterator begin_it, InputIterator end_it) {
        std::size_t assign_size = std::distance(begin_it, end_it);
        if (dims.size() < assign_size) {
            assign_size = dims.size();
            end_it = begin_it + assign_size;
        }
        std::copy(begin_it, end_it, dims.begin());
        size_n = assign_size;
    }

    inline S size() const {
        return size_n;
    }

    inline bool empty() const {
        return size_n == 0;
    }

    inline S & at(S n) {
        return dims[n];
    }

    inline S & operator[](S n) {
        return dims[n];
    }

    inline const S & at(S n) const {
        return dims[n];
    }

    inline const S & operator[](S n) const {
        return dims[n];
    }

    inline boost::array<S, 7>::const_iterator begin() const {
        return dims.begin();
    }

    inline boost::array<S, 7>::const_iterator end() const {
        return dims.begin() + size_n;
    }

    inline boost::array<S, 7>::iterator begin() {
        return dims.begin();
    }

    inline boost::array<S, 7>::iterator end() {
        return dims.begin() + size_n;
    }

    inline S get_total_dims_size() const {
        S total_dims_size = 1;
        for (boost::array<S, 7>::const_iterator it = begin(); it != end(); ++it) {
            if (*it) {
                total_dims_size *= (*it);
            } else {
                break;
            }
        }
        return total_dims_size;
    }

private:
    boost::array<S, 7> dims;
    std::size_t size_n;
};


typedef Dims Indexes;

/**
   @code
   std::vector<int> values(3 * 3 * 3) = {...};
   int begin = get_value(NDims<3>(3, 3, 3), NDims<3>(0, 0, 0), values); // => int begin = values[0][0][0];
   int end   = get_value(NDims<3>(3, 3, 3), NDims<3>(2, 2, 2), values); // => int end   = values[2][2][2];
   set_value(NDims<3>(3, 3, 3), NDims<3>(0, 0, 0), values, 123456);     // => values[0][0][0] = 123456;
   set_value(NDims<3>(3, 3, 3), NDims<3>(2, 2, 2), values, 987654);     // => values[2][2][2] = 987654;
   @endcode
 *
 */
template<std::size_t N>
class NDims {
public:
    typedef std::size_t S;
    typedef boost::array<S, N> DimArray;

    NDims() : dims() {}

    NDims(S a)
      : dims(make_array<S, 1>(a)) {}
    NDims(S a, S b) 
      : dims(make_array<S, 2>(a, b)) {}
    NDims(S a, S b, S c)
      : dims(make_array<S, 3>(a, b, c)) {}
    NDims(S a, S b, S c, S d)
      : dims(make_array<S, 4>(a, b, c, d)) {}
    NDims(S a, S b, S c, S d, S e)
      : dims(make_array<S, 5>(a, b, c, d, e)) {}
    NDims(S a, S b, S c, S d, S e, S f)
      : dims(make_array<S, 6>(a, b, c, d, e, f)) {}
    NDims(S a, S b, S c, S d, S e, S f, S g)
      : dims(make_array<S, 7>(a, b, c, d, e, f, g)) {}

    template <class Container>
    void assign(Container &values) {
        for (size_t i = 0; i < values.size(); ++i) {
            if (i >= dims.size()) break;
            dims[i] = static_cast<S>(values[i]);
        }
    }

    template <class InputIterator>
    void assign(InputIterator begin_it, InputIterator end_it) {
        std::size_t assign_size = std::distance(begin_it, end_it);
        if (dims.size() < assign_size) {
            assign_size = dims.size();
            end_it = begin_it + assign_size;
        }
        std::copy(begin_it, end_it, dims.begin());
    }

    inline S size() const {
        return dims.size();
    }

    inline bool empty() const {
        return dims.empty();
    }

    inline S & at(S n) {
        return dims[n];
    }

    inline S & operator[](S n) {
        return dims[n];
    }

    inline const S & at(S n) const {
        return dims[n];
    }

    inline const S & operator[](S n) const {
        return dims[n];
    }

    inline typename DimArray::const_iterator begin() const {
        return dims.begin();
    }

    inline typename DimArray::const_iterator end() const {
        return dims.end();
    }

    inline typename DimArray::iterator begin() {
        return dims.begin();
    }

    inline typename DimArray::iterator end() {
        return dims.end();
    }

    inline S get_total_dims_size() const {
        S total_dims_size = 1;
        for (typename DimArray::const_iterator it = begin(); 
            it != end(); ++it) {
            if (*it) {total_dims_size *= (*it);}
        }
        return total_dims_size;
    }

private:
    DimArray dims;
};

} // namespace Faster

} // namespace Array

#endif // ifndef

