/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK_hdf.h                                                */
/*  Abstract     : I/O toolkit header file                                   */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
#ifndef __SGTK_HDF_H__
#define __SGTK_HDF_H__

#include <string.h>
#include <math.h>
#include <typeinfo>
#include <vector>
#include "H5Cpp.h"
#include "SGTK.h"
#include "SGTK_ret_value.h"
#include "SGTK_dims.h"
#include "SGTK_cnf.h"

const std::string   ATTR_OFFSET("Offset");
const std::string   ATTR_SLOPE("Slope");
const std::string   ATTR_OFFSET_REF("Offset_reflectance");
const std::string   ATTR_SLOPE_REF("Slope_reflectance");
const std::string   ATTR_ERROR_DN("Error_DN");
const std::string   ATTR_ERROR_VALUE("Error_value");
const std::string   ATTR_DATA_INTERVAL_PIXEL("Data_interval_pixel");
const std::string   ATTR_DATA_INTERVAL_LINE("Data_interval_line");
const std::string   ATTR_RESAMPLING_INTERVAL("Resampling_interval");

enum RETVAL {
    RST_OK=0,
    RST_NONE=-1,
    RST_NG=-10
};

enum CompLevel {
    NONE = 0,
    LV1  = 1,
    LV2  = 2,
    LV3  = 3,
    LV4  = 4,
    LV5  = 5,
    LV6  = 6,
    LV7  = 7,
    LV8  = 8,
    LV9  = 9
};

enum BACKING_STORE {
    BACKING_STORE_OFF = 0,
    BACKING_STORE_ON  = 1
};

const hsize_t MAX_DIMS=6;

    
struct PrdAttr {
    void                 *p_value;
    std::vector<hsize_t> dims;
    std::string          name;
    hid_t                type;
    int                  rank;
    int                  length;
    H5T_str_t            strpad;
    PrdAttr() : p_value(NULL), dims(), name(), type(0), rank(0), length(0), strpad(H5T_STR_NULLPAD) {}
};

struct PrdGroupInfo {
    std::string             path;
    std::vector<PrdAttr>    attr;
    PrdGroupInfo() : path(), attr(0) {}
};

struct PrdDatasetInfo {
    std::vector<hsize_t>        cdims;
    std::vector<PrdAttr>        attr;
    std::string                 path;
    hid_t                       type;
    int                         rank;
    std::vector<hsize_t>        dims;
    void                        *p_value;
    unsigned int                level;
    std::vector<std::string>    label;
    int                         length;
    PrdDatasetInfo() : cdims(), attr(), path(), type(0), rank(0), dims(), p_value(NULL), level(0), length(0) {}
};

struct PrdAttrInfo {
    std::string             path;
    std::vector<PrdAttr>    attr;
    PrdAttrInfo() : path(), attr(0) {}
};

struct SGTKDims {
    hsize_t dim[MAX_DIMS];
    SGTKDims() : dim() {}
};

/**
 * convert endian with sizes.
 *
 * @param [in,out] area  source data.
 * @param [in]     size  size of source data.
 * @return destination data.
 *
   @code
   char data[] = "ABCD";
   reverse_endian(data, sizeof(data));
   @endcode
 */
template <class T>
inline void reverse_endian(T *area, std::size_t size) {
    unsigned char *target = reinterpret_cast<unsigned char *>(area);
    std::reverse(target, target + size);
}

/**
 * convert endian with type.
 *
 * @param [in] value source data.
 * @return destination data.
 *
   @code
   int num = 0x1234;
   int reverse_num = reverse_endian(num);
   @endcode
 */
template<class T>
inline T reverse_endian(const T &value) {
    if (sizeof(T) == 1) return value;
    T dest = value;
    char volatile *lsb = reinterpret_cast<char volatile*>(&dest);
    char volatile *msb = lsb + sizeof(T);
    std::reverse(lsb, msb);
    return dest;
}

/**
 * convert std::vector to pointer.
 *
 * @param [in]      src         source std::vector data.
 * @param [in/out]  dst         destination data pointer.
 * @param [in]      size        size of source data.
 * @param [in]      byte_order  byte order of target data.
 */
template<class T>
inline void vec2ptr(const std::vector<T> &src, T *dst, const size_t &size, const bool &byte_order) {
    if(byte_order==true) {
        for (size_t cnt = 0; cnt < size; ++cnt)
        {
            dst[cnt] = src.at(cnt);
        }
    } else {
        for (size_t cnt = 0; cnt < size; ++cnt)
        {
            dst[cnt] = reverse_endian(src.at(cnt));
        }
    }
}

/**
 * convert pointer to std::vector.
 *
 * @param [in]      src         source data pointer.
 * @param [in/out]  dst         destination std::vector data.
 * @param [in]      size        size of source data.
 * @param [in]      byte_order  byte order of target data.
 */
template<class T>
inline void ptr2vec(const T *src, std::vector<T> &dst, const size_t &size, const bool &byte_order) {
    if(byte_order==true) {
        for (size_t cnt = 0; cnt<size; ++cnt) { 
            dst.at(cnt) = *(src+cnt);
        }
    } else {
        for (size_t cnt = 0; cnt<size; ++cnt) { 
            dst.at(cnt) = reverse_endian(*(src+cnt));
        }
    }
}

/** 
 * linear interpolation.
 * @param [in]      x1      x1
 * @param [in]      y1      y1
 * @param [in]      x2      x2
 * @param [in]      y2      y2
 * @param [in]      x       x
 * @retrun y
 */
template<class T1, class T2>
inline static T2 linear_complement(T1 x1, T2 y1, T1 x2, T2 y2, T1 x) {
    return static_cast<T2>(((y1 - y2)/(x1 - x2)) * x + y2 - ((y1 - y2)/(x1 - x2)) * x2);
}

/** 
 * linear interpolation for angle data.
 * @param [in]      x1      x1
 * @param [in]      y1      y1
 * @param [in]      x2      x2
 * @param [in]      y2      y2
 * @param [in]      x       x
 * @retrun y
 */
template<class T1, class T2>
inline T2 linear_complement_angle(T1 x1, T2 y1, T1 x2, T2 y2, T1 x) {

    if (y1 < -360.0 || 360.0 < y1)  return y1;
    if (y2 < -360.0 || 360.0 < y2)  return y2;

    if (y1 > 90.0 && y2 < -90.0) y2 += 360.0;
    if (y2 > 90.0 && y1 < -90.0) y1 += 360.0;
    T2 result = static_cast<T2>(((y1 - y2)/(x1 - x2)) * x + y2 - ((y1 - y2)/(x1 - x2)) * x2);
    if (result >  180.0) result -= 360;
    if (result < -180.0) result += 360;
    return result;
}

int get_byteorder(hid_t id_file);
herr_t op_func(hid_t loc_id, const char *name, const H5O_info_t *info, void *operator_data);
int get_lines_par_scan(char *granule_id, const std::string &path);
hid_t get_strtype(const int size, const H5T_str_t strpad);
hsize_t get_type_size(hid_t id_file, const std::string path, const std::string name);
H5T_str_t get_strpad(hid_t id_file, const std::string path, const std::string name);
int get_resolution(hid_t id_file, const std::string &path);
int get_highest_resolution(hid_t id_file);
int get_max_line_pixel(hid_t id_file, const std::string &path, const std::string &name);
int get_max_line(hid_t id_file, const std::string &path);
int get_max_pixel(hid_t id_file, const std::string &path);
int get_max_line_pixel_raw_data(hid_t id_file, const std::string &path, const std::string &name);
int get_max_line_raw_data(hid_t id_file, const std::string &path);
int get_max_pixel_raw_data(hid_t id_file, const std::string &path);

class HDF {

friend class TestFixture;

public:
    static hid_t open_file(const std::string &filename, uint mode);

    static int close_file(hid_t id_file);

    static int set_group(hid_t id_file, const struct PrdGroupInfo &group);

    static int set_dataset(hid_t id_file, const struct PrdDatasetInfo &dataset);

    static void set_group_attribute(hid_t id_file, const struct PrdAttrInfo &attr);

    static void set_dataset_attribute(hid_t id_file, const struct PrdAttrInfo &attr);

    static void set_attribute(hid_t id_file, const struct PrdAttrInfo &attr, const H5G_obj_t objtype);

    static void get_dataset(hid_t id_file, struct PrdDatasetInfo *p_dataset);

    static void get_group_attribute(hid_t id_file, struct PrdAttrInfo *p_attr);

    static void get_dataset_attribute(hid_t id_file, struct PrdAttrInfo *p_attr);

    static void get_attribute(hid_t id_file, struct PrdAttrInfo *p_attr, const H5G_obj_t objtype);

    static int get_ds_dims(hid_t id_file, const std::string &path, int *p_rank, std::vector<hsize_t> *p_dims);

    static hsize_t get_group_attr_dims(hid_t id_file, const std::string &path, const std::string &name, int *p_rank, std::vector<hsize_t> *p_dims);

    static hsize_t get_dataset_attr_dims(hid_t id_file, const std::string &path, const std::string &name, int *p_rank, std::vector<hsize_t> *p_dims);

    static hsize_t get_attr_dims(hid_t id_file, const std::string &path, const H5G_obj_t objtype, const std::string &name, int *p_rank, std::vector<hsize_t> *p_dims);

    static hid_t get_type_id(hid_t id_file, const std::string &path);

    static bool exist_item(hid_t id_file, const std::string &path);

    static bool exist_group(hid_t id_file, const std::string &path);

    static bool exist_item_attributes(hid_t id_file, const std::string &path, const std::string &name);


    template <class T>
    static int get_dataset_numeric_value(hid_t id_file, const std::string &path, std::vector<hsize_t> &dims, std::vector<T> &value) {

        try {
            PrdDatasetInfo dts; 
            dts.path = path.c_str();
            value.resize(1);
            dts.type = get_typeid(value.at(0));
            int rc = get_ds_dims(id_file, dts.path, &dts.rank, &dts.dims);
            if (rc < SUCCESS){
                return ERROR_GET_DS_DIMS;
            }
            // Get elements size.
            size_t size=1;
            dims.resize(dts.rank);
            for(int i=0; i<dts.rank; i++) {
                dims[i] = dts.dims[i];
                size *= dts.dims[i];
            }
            value.resize(size);
            dts.p_value = &value.at(0);
            get_dataset(id_file, &dts);
        }
        catch( const H5::Exception &ex ) {
            return ERROR_GET_DATASET_NUMERIC_VALUE;
        }
        return SUCCESS;
    }

    static int get_float_value_with_slope_and_offset(hid_t id_file, const std::string &path, std::vector<hsize_t> &dims, std::vector<float> &value, const int type);
    static int get_double_value_with_slope_and_offset(hid_t id_file, const std::string &path, std::vector<hsize_t> &dims, std::vector<double> &value, const int type);

    static float get_slope(hid_t id_file, const std::string &path);

    static float get_offset(hid_t id_file, const std::string &path);

    static int get_dataset_attr_float(hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<float> &value);

    static float get_slope_reflectance(hid_t id_file, const std::string &path);

    static float get_offset_reflectance(hid_t id_file, const std::string &path);

    static float get_slope_radiance(hid_t id_file, const std::string &path);

    static float get_offset_radiance(hid_t id_file, const std::string &path);

    template <class Tvalue, class Tslope, class Toffset>
    static float get_mod_value(const Tvalue value, const Tslope slope, const Toffset offset) {
            return (float)value * (float)slope + (float)offset;
    }

    static int get_lat_lon_with_interval(hid_t id_file, const std::string &path, std::vector<int> &interval, int type, std::vector<hsize_t> &dims, std::vector<float> &value);

    template <class T>
    static int get_line_pixel_value_with_interval(
        hid_t id_file, const std::string &path, std::vector<int> &interval, std::vector<hsize_t> &dims, std::vector<T> &value, int type) {

        // Get dataset value.
        std::vector<T> temp;
        std::vector<T> get_value;
        std::vector<hsize_t> tmpdims;
        int rc = get_dataset_numeric_value(id_file, path, tmpdims, get_value);
        if (rc < SUCCESS){
            return rc;
        }

        // Get slope and offset value.
        float slope  = 0.0;
        float offset = 0.0;
        if(0==type) {
            slope  = get_slope(id_file, path);
            offset = get_offset(id_file, path);
        } else {
            slope  = get_slope_reflectance(id_file, path);
            offset = get_offset_reflectance(id_file, path);
        }
        if (slope <= 0 || offset < 0) {
            slope = 1;
            offset = 0;
        }

        // Get Mix value
        float vmin = 0;
        bool minret = get_min_attr_value(id_file, path, vmin);
        // Get Max value
        float vmax = 0;
        bool maxret = get_max_attr_value(id_file, path, vmax);
        // Get error value.
        float err = 0;
        bool errret = get_error_value(id_file, path, err);

        // flag set
        bool range_flag = true;
        if (minret == true && maxret == true && errret == true) {
            range_flag = true;
        } else {
            range_flag = false;
        }

        // Get Base
        T vbase = 0;
        bool base_flag = get_base_attr_value(id_file, path, vbase);

        // Get dataset value.
        temp.resize(get_value.size());
        for(int i=0; i<(int)get_value.size(); i++) {
            if (range_flag == true){
                if (get_value[i] < vmin || vmax < get_value[i]){
                    temp[i] = err;
                    continue;
                }
            }
            temp[i] = get_mod_value(get_value[i], slope, offset);

            if (base_flag == true){
                float out_base = 0;
                set_log_value(temp[i], vbase, out_base);
                temp[i] = out_base;
            }
        }

        // Get interval value.
        hsize_t interval_line  = (hsize_t)interval[0];
        hsize_t interval_pixel = (hsize_t)interval[1];
        // Get dataset (with interval) value.
        hsize_t allpixel = 1;
        std::vector<T> pre_result;
        std::vector<hsize_t> intdims(3);
        hsize_t maxline=0, maxpixel=0;
        if(tmpdims.size()==2) {
            intdims[0] = 1;
            intdims[1] = tmpdims[0];
            intdims[2] = tmpdims[1];
            maxline = get_max_line_raw_data(id_file, path);
            if(0==maxline) {
                maxline = get_max_line(id_file, path);
                if(0==maxline) {
                    maxline = (tmpdims[0])*interval_line-interval_line;
                }
            }
            maxpixel = get_max_pixel_raw_data(id_file, path);
            if(0==maxpixel) {
                maxpixel = get_max_pixel(id_file, path);
                if(0==maxpixel) {
                    maxpixel = (tmpdims[1])*interval_pixel-interval_pixel;
                }
            }
            dims.push_back(maxline);
            dims.push_back(maxpixel);
        } else 
        if(tmpdims.size()==3) {
            intdims[0] = tmpdims[0];
            intdims[1] = tmpdims[1];
            intdims[2] = tmpdims[2];
            maxline = get_max_line_raw_data(id_file, path);
            if(0==maxline) {
                maxline = get_max_line(id_file, path);
                if(0==maxline) {
                    maxline = (tmpdims[1])*interval_line-interval_line;
                }
            }
            maxpixel = get_max_pixel_raw_data(id_file, path);
            if(0==maxpixel) {
                maxpixel = get_max_pixel(id_file, path);
                if(0==maxpixel) {
                    maxpixel = (tmpdims[2])*interval_pixel-interval_pixel;
                }
            }
            dims.push_back(tmpdims[0]);
            dims.push_back(maxline);
            dims.push_back(maxpixel);
        }
        for(size_t i=0L; i<dims.size(); ++i) allpixel *= dims.at(i);

        int idx1=0, idx2=0;
        double x1=0.0, y1=0.0, x2=0.0, y2=0.0, x=0.0, y=0.0;
        unsigned int line = 0, pixel = 0;

        hsize_t dim_pixel = 1;
        if(intdims[2] == 1) { // When the element size is 1 pixel, interpolation is not executed.
            pre_result = temp;
        } else {
            // Expand pixel.
            dim_pixel = maxpixel;
            for(hsize_t l=0L; l<intdims[0]; ++l) {
                for(hsize_t i=0L; i<intdims[1]; ++i) {
                    for(hsize_t j=0L; j<intdims[2]; ++j) {
                        x1 = (double)j;
                        x2 = (double)j + (double)interval_pixel;
                        for(hsize_t k=0L; k<interval_pixel; ++k) {
                            if((j*interval_pixel+k)>=dim_pixel) {
                            } else if(0==k) {
                                pre_result.push_back(temp[(l*intdims[1]*intdims[2])+(i*intdims[2])+x1]);
                            } else if( (intdims[2]-1)<=j ) {
                                T diff = pre_result.at(pre_result.size()-1) - pre_result.at(pre_result.size()-2);
                                pre_result.push_back( pre_result.at(pre_result.size()-1) + diff);
                            } else {
                                idx1 = (int)( (l*intdims[1]*intdims[2])+(i*intdims[2])+ x1 );
                                idx2 = (int)( (l*intdims[1]*intdims[2])+(i*intdims[2])+(x1+1) );
                                y1 = temp[idx1];
                                y2 = temp[idx2];
                                x  = x1 + k;
                                y = linear_complement(x1, y1, x2, y2, x); 
                                pre_result.push_back((T)y);
                            }
                            ++pixel;
                        }
                    }
                }
            }
        }

        // Expand line.
        for(hsize_t l=0L; l<intdims[0]; ++l) {
            line = 0;
            for(hsize_t i=0L; i<intdims[1]; ++i) {
                x1 = (double)i;
                x2 = (double)i + (double)interval_line;
                for(hsize_t k=0; k<interval_line; ++k) {
                    for(hsize_t j=0L; j<dim_pixel; ++j) {
                        if(0==k) {
                            value.push_back(pre_result[(l*intdims[1]*dim_pixel)+(x1*dim_pixel)+j]);
                            if(value.size()>=allpixel) return SUCCESS;
                        } else if( (intdims[1]-1)<=i ) {
                            int pos1 = (value.size())-(dim_pixel*1);
                            int pos2 = (value.size())-(dim_pixel*2);
                            T diff = value.at(pos1) - value.at(pos2);

                            value.push_back( value.at(pos1) + diff);
                            if(value.size()>=allpixel) return SUCCESS;
                        } else {
                            idx1 = (int)( (l*intdims[1]*dim_pixel)+( x1   *dim_pixel)+j );
                            idx2 = (int)( (l*intdims[1]*dim_pixel)+((x1+1)*dim_pixel)+j );
                            y1 = pre_result[idx1];
                            y2 = pre_result[idx2];
                            x  = x1 + k;
                            y = linear_complement(x1, y1, x2, y2, x); 
                            value.push_back((T)y);
                            if(value.size()>=allpixel) return SUCCESS;
                        }
                    }
                    ++line;
                    if(line>=maxline) { k = interval_line; i = intdims[1]; }
                }
            }
        }
        return SUCCESS;
    }

    static int get_data_interval_pixel(hid_t id_file, const std::string &path);

    static int get_data_interval_line(hid_t id_file, const std::string &path);

    static int get_resampling_interval(hid_t id_file, const std::string &path);

    static void get_dataset_attr_int(hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<int> &value);

    static int set_float_value_without_slope_and_offset(
        hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<float> &value, const int type,
        const CompLevel cmplvl, const int chunk);

    static int set_double_value_without_slope_and_offset(
        hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<double> &value, const int type,
        const CompLevel cmplvl, const int chunk);

    template <class Tvalue, class Tslope, class Toffset>
    static float get_org_value(const Tvalue value, const Tslope slope, const Toffset offset) {
            return ((float)value - (float)offset) / (float)slope;
    }

    static int set_lat_lon_without_interval(
        hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<float> &value, int type);

    template <class T>
    static int set_angle_value_without_interval_2d(
        hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<T> &value, const std::vector<int> &interval,
        const CompLevel cmplvl, const int chunk, int type) {

        std::vector<T>  cull_value;

        int idx1=0, idx2=0;
        double x1=0.0, y1=0.0, x2=0.0, y2=0.0, x=0.0, y=0.0;

        hsize_t gridnum[3] = { 0 };
        gridnum[0] = ceil((double)(dims.at(0)-1)/(double)interval.at(0))+1;
        gridnum[1] = ceil((double)(dims.at(1)-1)/(double)interval.at(1))+1;

        for(hsize_t i=0, i2=0; i2<gridnum[0]; i+=interval[0], ++i2) {
            for(hsize_t j=0, j2=0; j2<gridnum[1]; j+=interval[1], ++j2) {
                Array::Faster::Dims ddims(dims[0], dims[1]);
                if( (i<dims[0]) && (j>=dims[1]) ) { // interpolation of column
                    if (((int)dims[1] - 1) - interval[1] < 0) {
                        x1 = 0.0;
                    } else {
                        x1 = (dims[1] - 1) - interval[1];
                    }
                    x2 = dims[1] - 1;
                    idx1 = Array::Faster::calc_index( ddims, Array::Faster::Indexes(i, x1) );
                    idx2 = Array::Faster::calc_index( ddims, Array::Faster::Indexes(i, x2) );
                    y1 = value[idx1];
                    y2 = value[idx2];
                    x  = j;
                    y = linear_complement_angle(x1, y1, x2, y2, x); 
                    cull_value.push_back((T)y);
                }
                else
                if( (i>=dims[0]) && (j<dims[1]) ) { // interpolation of row
                    if (((int)dims[0] - 1) - interval[0] < 0) {
                        x1 = 0.0;
                    } else {
                        x1 = (dims[0] - 1) - interval[0];
                    }
                    x2 = dims[0] - 1;
                    idx1 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(x1, j));
                    idx2 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(x2, j));
                    y1 = value[idx1];
                    y2 = value[idx2];
                    x  = i;
                    y = linear_complement_angle(x1, y1, x2, y2, x); 
                    cull_value.push_back((T)y);
                }
                else
                if( (i>=dims[0]) && (j>=dims[1]) ) { // interpolation of row and column
                    x1 = j-(interval[1]*2);
                    x2 = j-interval[1];
                    y1 = cull_value.at(cull_value.size()-2);
                    y2 = cull_value.at(cull_value.size()-1);
                    x  = j;
                    y = linear_complement_angle(x1, y1, x2, y2, x); 
                    cull_value.push_back((T)y);
                }
                else { // do not interpolation
                    size_t index = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, j));
                    cull_value.push_back(value[index]);
                }
            }
        }

        std::vector<hsize_t> dim(dims.size());
        for(unsigned int i=0; i<dims.size(); ++i) {
            dim[i] = gridnum[i];
        }

        std::vector<int> chunks;
        if(NONE!=cmplvl) {
            for(int i=0; i<(int)dims.size(); ++i) chunks.push_back(chunk);
        }

        std::vector<T>  set_value;
        // Get slope and offset value.
        float slope  = 0.0;
        float offset = 0.0;
        if(0==type) {
            slope  = get_slope(id_file, path);
            offset = get_offset(id_file, path);
        } else {
            slope  = get_slope_reflectance(id_file, path);
            offset = get_offset_reflectance(id_file, path);
        }
        if (slope <= 0 || offset < 0) {
            slope = 1;
            offset = 0;
        }

        // Get Mix value
        float vmin = 0;
        bool minret = get_min_attr_value(id_file, path, vmin);
        // Get Max value
        float vmax = 0;
        bool maxret = get_max_attr_value(id_file, path, vmax);
        // Get error value.
        float err = 0;
        bool errret = get_error_value(id_file, path, err);

        // flag set
        bool range_flag = true;
        if (minret == true && maxret == true && errret == true) {
            range_flag = true;
        } else {
            range_flag = false;
        }

        // Get Base
        float vbase = 0;
        bool base_flag = get_base_attr_value(id_file, path, vbase);

        for(int i=0; i<(int)cull_value.size(); i++) {
            if (base_flag == true){
                set_pow_value(cull_value[i], vbase, cull_value[i]);
            }
            set_value.push_back(get_org_value(cull_value[i], slope, offset));
            if (range_flag == true){
                if (set_value[i] < vmin || vmax < set_value[i]){
                    set_value[i] = err;
                }
            }
        }

        int rc = set_dataset_numeric_value(id_file, path, dim, cull_value, cmplvl, chunks);
        return rc;
    }

    template <class T>
    static int set_angle_value_without_interval_3d(
        hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<T> &value, const std::vector<int> &interval,
        const CompLevel cmplvl, const int chunk, int type) {

        std::vector<T>  cull_value;

        int idx1=0, idx2=0;
        double x1=0.0, y1=0.0, x2=0.0, y2=0.0, x=0.0, y=0.0;

        hsize_t gridnum[3] = { 0 };
        gridnum[0] = dims[0];
        gridnum[1] = ceil((double)(dims.at(1)-1)/(double)interval.at(1))+1;
        gridnum[2] = ceil((double)(dims.at(2)-1)/(double)interval.at(2))+1;

        for(hsize_t i=0; i<dims[0]; i+=interval[0]) {
            for(hsize_t j=0, j2=0; j2<gridnum[1]; j+=interval[1], ++j2) {
                for(hsize_t k=0, k2=0; k2<gridnum[2]; k+=interval[2], ++k2) {
                    Array::Faster::Dims ddims(dims[0], dims[1], dims[2]);
                    if( (j<dims[1]) && (k>=dims[2]) ) { // intepolation of column
                        if (((int)dims[2] - 1) - interval[2] < 0) {
                            x1 = 0.0;
                        } else {
                            x1 = (dims[2] - 1) - interval[2];
                        }
                        x2 = dims[2] - 1;
                        idx1 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, j, x1));
                        idx2 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, j, x2));
                        y1 = value[idx1];
                        y2 = value[idx2];
                        x  = k;
                        y = linear_complement_angle(x1, y1, x2, y2, x); 
                        cull_value.push_back((T)y);
                    }
                    else
                    if( (j>=dims[1]) && (k<dims[2]) ) { // interpolation of row
                        if (((int)dims[1] - 1) - interval[1] < 0) {
                            x1 = 0.0;
                        } else {
                            x1 = (dims[1] - 1) - interval[1];
                        }
                        x2 = dims[1] - 1;
                        idx1 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, x1, k));
                        idx2 = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, x2, k));
                        y1 = value[idx1];
                        y2 = value[idx2];
                        x  = j;
                        y = linear_complement_angle(x1, y1, x2, y2, x); 
                        cull_value.push_back((T)y);
                    } 
                    else
                    if( (j>=dims[1]) && (k>=dims[2]) ) {    // interpolation of row and column
                        x1 = k-(interval[2]*2);
                        x2 = k-interval[2];
                        y1 = cull_value.at(cull_value.size()-2);
                        y2 = cull_value.at(cull_value.size()-1);
                        x  = k;
                        y = linear_complement_angle(x1, y1, x2, y2, x); 
                        cull_value.push_back((T)y);
                    }
                    else {  // do not interpolation
                        size_t index = Array::Faster::calc_index(ddims, Array::Faster::Indexes(i, j, k));
                        cull_value.push_back(value[index]);
                    }
                }
            }
        }

        std::vector<hsize_t> dim(dims.size());
        for(unsigned int i=0; i<dims.size(); ++i) {
            dim[i] = gridnum[i];
        }

        std::vector<int> chunks;
        if(NONE!=cmplvl) {
            for(int i=0; i<(int)dims.size(); ++i) chunks.push_back(chunk);
        }

        std::vector<T>  set_value;
        // Get slope and offset value.
        float slope  = 0.0;
        float offset = 0.0;
        if(0==type) {
            slope  = get_slope(id_file, path);
            offset = get_offset(id_file, path);
        } else {
            slope  = get_slope_reflectance(id_file, path);
            offset = get_offset_reflectance(id_file, path);
        }
        if (slope <= 0 || offset < 0) {
            slope = 1;
            offset = 0;
        }

        // Get Mix value
        float vmin = 0;
        bool minret = get_min_attr_value(id_file, path, vmin);
        // Get Max value
        float vmax = 0;
        bool maxret = get_max_attr_value(id_file, path, vmax);
        // Get error value.
        float err = 0;
        bool errret = get_error_value(id_file, path, err);

        // flag set
        bool range_flag = true;
        if (minret == true && maxret == true && errret == true) {
            range_flag = true;
        } else {
            range_flag = false;
        }

        // Get Base
        float vbase = 0;
        bool base_flag = get_base_attr_value(id_file, path, vbase);

        for(int i=0; i<(int)cull_value.size(); i++) {
            if (base_flag == true){
                set_pow_value(cull_value[i], vbase, cull_value[i]);
            }
            set_value.push_back(get_org_value(cull_value[i], slope, offset));
            if (range_flag == true){
                if (set_value[i] < vmin || vmax < set_value[i]){
                    set_value[i] = err;
                }
            }
        }

        int rc = set_dataset_numeric_value(id_file, path, dim, cull_value, cmplvl, chunks);
        return rc;
    }

    static int get_group_attr_string_value(
            hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<std::string> &value);

    static int get_dataset_attr_string_value(
            hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<std::string> &value);

    static int set_group_attr_string_value(
            hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<std::string> &value);

    static int set_dataset_attr_string_value(
            hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<std::string> &value);

    template <class T>
    static int get_group_attr_numeric_value(
            hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<T> &value) {

            try{
                PrdAttrInfo attrinfo;
                attrinfo.path = path.c_str();
                // Set attribute info.
                PrdAttr attr;
                attr.name = name.c_str();
                value.resize(1);
                attr.type = get_typeid(value.at(0));
                int rc = get_group_attr_dims(id_file, attrinfo.path, attr.name, &attr.rank, &attr.dims);
                if (rc < SUCCESS){
                    return rc;
                }
                // Get elements size.
                size_t size=1;
                dims.resize(attr.rank);
                for(int i=0; i<attr.rank; i++) {
                    dims[i] = attr.dims[i];
                    size *= attr.dims[i];
                }
                value.resize(size);
                attr.p_value = &value.at(0);
                attrinfo.attr.push_back(attr);
                get_group_attribute(id_file, &attrinfo);
            }
            catch( const H5::Exception &ex ) {
                return ERROR_GET_GROUP_ATTR;
            }
            return SUCCESS;
   }

    template <class T>
    static int get_dataset_attr_numeric_value(
            hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<T> &value) {

            try{
                PrdAttrInfo attrinfo;
                attrinfo.path = path.c_str();
                // Set attribute info.
                PrdAttr attr;
                attr.name = name.c_str();
                value.resize(1);
                attr.type = get_typeid(value.at(0));
                int rc = get_dataset_attr_dims(id_file, attrinfo.path, attr.name, &attr.rank, &attr.dims);
                if (rc < SUCCESS){
                    return rc;
                }
                // Get elements size.
                size_t size=1;
                dims.resize(attr.rank);
                for(int i=0; i<attr.rank; i++) {
                    dims[i] = attr.dims[i];
                    size *= attr.dims[i];
                }
                value.resize(size);
                attr.p_value = &value.at(0);
                attrinfo.attr.push_back(attr);
                get_dataset_attribute(id_file, &attrinfo);
            }
            catch( const H5::Exception &ex ) {
                return ERROR_GET_DATASET_ATTR;
            }
            return SUCCESS;
    }

    template <class T>
    static int set_group_attr_numeric_value(
            hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<T> &value) {

            try{
                PrdAttrInfo attrinfo;
                attrinfo.path = path.c_str();
                // Set attribute info.
                PrdAttr attr;
                attr.name = name.c_str();
                attr.type = get_typeid(value.at(0));
                attr.rank = dims.size();
                size_t size=1;
                // Set dimension size.
                for(int i=0; i<attr.rank; i++) {
                    attr.dims.push_back(dims[i]);
                    size *= dims[i];
                }
            
                T *p_value = const_cast<T*>(&value.at(0));
                attr.p_value = p_value;
                attrinfo.attr.push_back(attr);

                // Set group info.
                PrdGroupInfo groupinfo;
                groupinfo.path = attrinfo.path;
                groupinfo.attr = attrinfo.attr;
                int rc = set_group(id_file, groupinfo);
                if (rc < SUCCESS){
                    return rc;
                }

                set_group_attribute(id_file, attrinfo);
            }
            catch( const H5::Exception &ex ) {
               return ERROR_SET_GROUP_ATTR;
            }
            return SUCCESS;
    }

    template <class T>
    static int set_dataset_attr_numeric_value(
            hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<T> &value) {

            try{
                PrdAttrInfo attrinfo;
                attrinfo.path = path.c_str();
                // Set attribute info.
                PrdAttr attr;
                attr.name = name.c_str();
                attr.type = get_typeid(value.at(0));
                attr.rank = dims.size();
                size_t size=1;
                // Set dimension size.
                for(int i=0; i<attr.rank; i++) {
                    attr.dims.push_back(dims[i]);
                    size *= dims[i];
                }

                T *p_value = const_cast<T*>(&value.at(0));
                attr.p_value = p_value;
                attrinfo.attr.push_back(attr);

                set_dataset_attribute(id_file, attrinfo);
            }
            catch( const H5::Exception &ex ) {
               return ERROR_SET_DATASET_ATTR;
            }
            return SUCCESS;
    }

    template <class T>
    static int set_dataset_numeric_value(
            hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<T> &value, 
            const CompLevel &cmplvl, const std::vector<int> &chunk) {

            PrdDatasetInfo dts; 
            dts.path = path;
            dts.type = get_typeid(value.at(0));
            dts.rank = dims.size();
            size_t size=1;
            // Set dimension size.
            for(int i=0; i<dts.rank; i++) {
                dts.dims.push_back(dims[i]);
                size *= dims[i];
            }
            // Set chunk size.
            for(int i=0; i<(int)chunk.size(); i++) {
                dts.cdims.push_back(chunk[i]);
            }

            T *p_value = const_cast<T*>(&value.at(0));
            dts.level = cmplvl;
            dts.p_value = p_value;

            int rc = set_dataset(id_file, dts);
            return rc;
    }

    template <class T>
    static hid_t get_typeid(const T value) {

hid_t typeid_tbl[2][11] = {
    {
        H5T_STD_I8LE,
        H5T_STD_U8LE,
        H5T_STD_I16LE,
        H5T_STD_U16LE,
        H5T_STD_I32LE,
        H5T_STD_U32LE,
        H5T_STD_I64LE,
        H5T_STD_U64LE,
        H5T_IEEE_F32LE,
        H5T_IEEE_F64LE,
        H5T_C_S1
    },
    {
        H5T_STD_I8BE,
        H5T_STD_U8BE,
        H5T_STD_I16BE,
        H5T_STD_U16BE,
        H5T_STD_I32BE,
        H5T_STD_U32BE,
        H5T_STD_I64BE,
        H5T_STD_U64BE,
        H5T_IEEE_F32BE,
        H5T_IEEE_F64BE,
        H5T_C_S1
    }
};

            hid_t result = 0;
            bool isLittle = conf_obj.byte_order;
            int idx=0;
            if(!isLittle) idx=1; 

            if(typeid(value)==typeid(char) || typeid(value)==typeid(int8_t)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_I8_TYPE];
            } else 
            if(typeid(value)==typeid(signed char)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_I8_TYPE];
            } else 
            if((typeid(value)==typeid(unsigned char))||(typeid(value)==typeid(uint8_t))) {
                result = typeid_tbl[idx][SGTK_H5T_STD_U8_TYPE];
            } else 
            if(typeid(value)==typeid(short)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_I16_TYPE];
            } else 
            if(typeid(value)==typeid(unsigned short)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_U16_TYPE];
            } else 
            if(typeid(value)==typeid(int)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_I32_TYPE];
            } else
            if(typeid(value)==typeid(unsigned int)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_U32_TYPE];
            } else
            if(typeid(value)==typeid(long)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_I64_TYPE];
            } else
            if(typeid(value)==typeid(unsigned long)) {
                result = typeid_tbl[idx][SGTK_H5T_STD_U64_TYPE];
            } else
            if(typeid(value)==typeid(float)) {
                result = typeid_tbl[idx][SGTK_H5T_IEEE_F32_TYPE];
            } else
            if(typeid(value)==typeid(double)) {
                result = typeid_tbl[idx][SGTK_H5T_IEEE_F64_TYPE];
            } else
            if(typeid(value)==typeid(std::string)) {
                result = typeid_tbl[idx][SGTK_H5T_C_S1_TYPE];
            }

            return result;
    }

    static void remove_attributes(hid_t id_file, const std::string &path, const std::string &name);

    static int set_parent_group(hid_t id_file, const std::string &path);

    static hid_t create_dataset(hid_t id_file, const PrdDatasetInfo &dataset, hid_t &r_memtype, hid_t &r_dataspace, hid_t &r_plist);

    static hid_t open_dataset(hid_t id_file, const PrdDatasetInfo &dataset, hid_t &r_memtype, hid_t &r_dataspace, hid_t &r_plist);

    static hid_t get_attribute_id(const hid_t id, const std::string &name, const hid_t type, const hid_t ds);

    static hid_t get_memtype(const hid_t type);

    static void set_chunk_plist(const struct PrdDatasetInfo &dataset, hid_t &r_plist);

    static void set_comp_plist(const struct PrdDatasetInfo &dataset, hid_t &r_plist);

    template <class T>
    static void set_log_value(const T &in, const T &base, T &out) {
        if(typeid(in)==typeid(float))   // float
        {
            out = logf(in) / logf(base);
        }
        else                            // double
        {
            out = log(in) / log(base);
        }
    }

    template <class T>
    static void set_pow_value(const T &in, const T &base, T &out) {
        if(typeid(in)==typeid(float))   // float
        {
            out = powf(base, in);
        }
        else                            // double
        {
            out = pow(base, in);
        }
    }


    template <class T>
    static bool get_error_value(hid_t id_file, const std::string &path, T &out) {
        std::vector<hsize_t> dims;
        std::vector<T> value;

        int elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.range_err_value);
        if (elment != -1)
        {
            int rc = get_dataset_attr_numeric_value(id_file, path, conf_obj.range_err_value.at(elment), dims, value);
            if (rc < SUCCESS){
                return false;
            }
            out = value[0];
        }
        else
        {
            return false;
        }
        return true;
    }

    template <class T>
    static bool get_min_attr_value(hid_t id_file, const std::string &path, T &out) {
        std::vector<hsize_t> dims;
        std::vector<T> value;

        int elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.range_min_attr_name);
        if (elment != -1)
        {
            int rc = get_dataset_attr_numeric_value(id_file, path, conf_obj.range_min_attr_name.at(elment), dims, value);
            if (rc < SUCCESS){
                return false;
            }
            out = value[0];
        }
        else
        {
            return false;
        }
        return true;
    }

    template <class T>
    static bool get_max_attr_value(hid_t id_file, const std::string &path, T &out) {
        std::vector<hsize_t> dims;
        std::vector<T> value;

        int elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.range_max_attr_name);
        if (elment != -1)
        {
            int rc = get_dataset_attr_numeric_value(id_file, path, conf_obj.range_max_attr_name.at(elment), dims, value);
            if (rc < SUCCESS){
                return false;
            }
            out = value[0];
        }
        else
        {
            return false;
        }
        return true;
    }

    template <class T>
    static bool get_base_attr_value(hid_t id_file, const std::string &path, T &out) {
        std::vector<hsize_t> dims;
        std::vector<T> value;

        int elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.correct_base_attr_name);
        if (elment != -1)
        {
            int rc = get_dataset_attr_numeric_value(id_file, path, conf_obj.correct_base_attr_name.at(elment), dims, value);
            if (rc < SUCCESS){
                return false;
            }
            out = value[0];
        }
        else
        {
            return false;
        }
        return true;
    }


protected:
private:
    HDF(const HDF&);

    HDF& operator = (const HDF&);

};

#endif    // __SGTK_HDF_H__
