/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK.cpp                                                  */
/*  Abstract     : I/O toolkit API function                                  */
/*  Function List                                                            */
/*      SGTK_open_file()           : HDF5 file open                          */
/*      SGTK_close_file()          : HDF5 file close                         */
/*      SGTK_make_group()          : HDF5 group create                       */
/*      SGTK_utc_to_tai93()        : Convert from UTC to TAI93               */
/*      SGTK_tai93_to_utc()        : Convert from TAI93 to UTC               */
/*      SGTK_malloc()              : Memory allocate                         */
/*      SGTK_free()                : Memory release                          */
/*      SGTK_get_global_latlon()   : Get latitude/longitude binary           */
/*      SGTK_load_config()         : Load configuration file                 */
/*      SGTK_get_dims()            : Get HDF5 daaset dimension size          */
/*      SGTK_get_dataset_int()     : Get HDF5 dataset integer value          */
/*      SGTK_get_dataset_int8()    : Get HDF5 dataset 8bit integer value     */
/*      SGTK_get_dataset_int16()   : Get HDF5 dataset 16bit integer value    */
/*      SGTK_get_dataset_int32()   : Get HDF5 dataset 32bit integer value    */
/*      SGTK_get_dataset_uint()    : Get HDF5 dataset integer value          */
/*      SGTK_get_dataset_uint8()   : Get HDF5 dataset 8bit integer value     */
/*      SGTK_get_dataset_uint16()  : Get HDF5 dataset 16bit integer value    */
/*      SGTK_get_dataset_uint32()  : Get HDF5 dataset 32bit integer value    */
/*      SGTK_get_dataset_float()   : Get HDF5 dataset float value            */
/*      SGTK_get_dataset_float32() : Get HDF5 dataset 32bit float value      */
/*      SGTK_get_dataset_float64() : Get HDF5 dataset 64bit double value     */
/*      SGTK_get_dataset_latlon()  : Get HDF5 dataset latitude/longitude     */
/*      SGTK_get_meta_string()     : Get HDF5 metadata string value          */
/*      SGTK_get_meta_int()        : Get HDF5 metadata integer value         */
/*      SGTK_get_meta_int8()       : Get HDF5 metadata 8bit integer value    */
/*      SGTK_get_meta_int16()      : Get HDF5 metadata 16bit integer value   */
/*      SGTK_get_meta_int32()      : Get HDF5 metadata 32bit integer value   */
/*      SGTK_get_meta_uint()       : Get HDF5 metadata integer value         */
/*      SGTK_get_meta_uint8()      : Get HDF5 metadata 8bit integer value    */
/*      SGTK_get_meta_uint16()     : Get HDF5 metadata 16bit integer value   */
/*      SGTK_get_meta_uint32()     : Get HDF5 metadata 32bit integer value   */
/*      SGTK_get_meta_float()      : Get HDF5 metadata float value           */
/*      SGTK_get_meta_float32()    : Get HDF5 metadata 32bit float value     */
/*      SGTK_get_meta_float64()    : Get HDF5 metadata 64bit double value    */
/*      SGTK_set_dims()            : Set HDF5 dataset dimension size         */
/*      SGTK_set_dataset_int()     : Set HDF5 dataset native integer value   */
/*      SGTK_set_dataset_int8()    : Set HDF5 dataset 8bit integer value     */
/*      SGTK_set_dataset_int16()   : Set HDF5 dataset 16bit integer value    */
/*      SGTK_set_dataset_int32()   : Set HDF5 dataset 32bit integer value    */
/*      SGTK_set_dataset_uint()    : Set HDF5 dataset native integer value   */
/*      SGTK_set_dataset_uint8()   : Set HDF5 dataset 8bit integer value     */
/*      SGTK_set_dataset_uint16()  : Set HDF5 dataset 16bit integer value    */
/*      SGTK_set_dataset_uint32()  : Set HDF5 dataset 32bit integer value    */
/*      SGTK_set_dataset_float()   : Set HDF5 dataset native float value     */
/*      SGTK_set_dataset_float32() : Set HDF5 dataset 32bit float value      */
/*      SGTK_set_dataset_float64() : Set HDF5 dataset 64bit double value     */
/*      SGTK_set_dataset_latlon()  : Set HDF5 dataset latitude/longitude     */
/*      SGTK_set_meta_string()     : Set HDF5 metadata string value          */
/*      SGTK_set_meta_int()        : Set HDF5 metadata native integer value  */
/*      SGTK_set_meta_int8()       : Set HDF5 metadata 8bit integer value    */
/*      SGTK_set_meta_int16()      : Set HDF5 metadata 16bit integer value   */
/*      SGTK_set_meta_int32()      : Set HDF5 metadata 32bit integer value   */
/*      SGTK_set_meta_uint()       : Set HDF5 metadata native integer value  */
/*      SGTK_set_meta_uint8()      : Set HDF5 metadata 8bit integer value    */
/*      SGTK_set_meta_uint16()     : Set HDF5 metadata 16bit integer value   */
/*      SGTK_set_meta_uint32()     : Set HDF5 metadata 32bit integer value   */
/*      SGTK_set_meta_float()      : Set HDF5 metadata native float value    */
/*      SGTK_set_meta_float32()    : Set HDF5 metadata 32bit float value     */
/*      SGTK_set_meta_float64()    : Set HDF5 metadata 64bit double value    */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
/*! \mainpage GCOM-C1 SGLI Product I/O Tool Kit
 *
 * \section project GCOM-C1 SGLI Product I/O Tool Kit
 * \subsection title Program Module Design
 * 2015/03/20.
 */
#include <sys/stat.h>
#include "SGTK.h"
#include "SGTK_hdf.h"
#include "SGTK_cnf.h"
#include "SGTK_date_time.h"
#include "SGTK_ret_value.h"
#include "SGTK_latlon.h"

struct LatLonStrct {
        float   lat;
        float   lon;
};

// O[oϐ
std::map<std::string, std::string> SGLI_ACCESS_LABEL_TABLE;

/** open hdf5 file.
 *
 * @param   file_name       [char*][in] Product file name.
 * @param   mode            [int][in]   file open mode.
 * @return  hdf5 file id.
 */
//hid_t SGTK_open_file(char *file_name, uint mode)
hid_t SGTK_open_file(char *file_name)
{
    std::string FILENAME(file_name);    
    return (int)HDF::open_file(FILENAME, SGTK_HDF_ACC_RDWR);
}

/** close hdf5 file.
 *
 * @param   HDF_file_id     [int][in]   hdf5 file id.
 * @return  result status.
 */
int SGTK_close_file(hid_t HDF_file_id)
{
    int result = SUCCESS;
    HDF::close_file(HDF_file_id);
    return result;
}

/** make group.
 *
 * @param   HDF_file_id     [hid_t][in]   hdf5 file id.
 * @param   group           [char*][in] group name.
 * @return  result status.
 */
int SGTK_make_group(hid_t HDF_file_id, char *group)
{
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (group == NULL)
    {
        return ERROR_PARAM;
    }

    return HDF::set_parent_group(HDF_file_id, group);
}

/** convert utc to tai93.
 *
 * @param   year            [int][in]       year.
 * @param   month           [int][in]       month.
 * @param   day             [int][in]       day.
 * @param   hour            [int][in]       hour.
 * @param   min             [int][in]       minutes.
 * @param   sec             [int][in]       second.
 * @param   msec            [int][in]       milli-second.
 * @param   tai93time       [double*][out]  TAI93 time.
 * @return  result status.
 */
int SGTK_utc_to_tai93(int year, int month, int day, int hour, int min, int sec, int msec, double *tai93time)
{

// data
    utc p_utc_time;

// file read
    int rc = com_get_leapsecond_formal(conf_obj.leap_sec_file_path.c_str());
    if (rc != SUCCESS) {
        return rc;
    }

// convert
    p_utc_time.year = year;
    p_utc_time.month = month;
    p_utc_time.day = day;
    p_utc_time.hour = hour;
    p_utc_time.minute = min;
    p_utc_time.second = sec;
    p_utc_time.microsecond = msec * 1000.0;

    rc = com_time_utc_tai93(&p_utc_time, tai93time);

// free
    com_release_leapsecond();

    if (rc != SUCCESS) {
        return rc;
    }

    return SUCCESS;

}

/** convert tai93 to utc.
 *
 * @param   tai93time       [double][in]    TAI93 time.
 * @param   year            [int*][out]     year.
 * @param   month           [int*][out]     month.
 * @param   day             [int*][out]     day.
 * @param   hour            [int*][out]     hour.
 * @param   min             [int*][out]     minutes.
 * @param   sec             [int*][out]     second.
 * @param   msec            [int*][out]     milli-second.
 * @return  result status.
 */
int SGTK_tai93_to_utc(double tai93time, int *year, int *month, int *day, int *hour, int *min, int *sec, int *msec)
{

// data
    utc p_utc_time;

// file read
    int rc = com_get_leapsecond_formal(conf_obj.leap_sec_file_path.c_str());
    if (rc != SUCCESS) {
        return rc;
    }

// convert
    rc = com_time_tai93_utc(tai93time, &p_utc_time);

// free
    com_release_leapsecond();

    if (rc != SUCCESS) {
        return rc;
    }

    *year  = p_utc_time.year;
    *month = p_utc_time.month;
    *day   = p_utc_time.day;
    *hour  = p_utc_time.hour;
    *min   = p_utc_time.minute;
    *sec   = p_utc_time.second;
    *msec  = p_utc_time.microsecond / 1000.0;

    return SUCCESS;

}

/** allocate memory.
 *
 * @param   dims    [int*][in]  number of elements in each dimensions.
 * @param   rank    [int][in]   number of dimensions.
 * @param   size    [int][in]   size of a elements.
 * @return  address of allocated memory.
 */
void *SGTK_malloc(int *dims, int rank, int size)
{

    int total_size = 1;

    if (dims == NULL){
        return NULL;
    }
    else if (rank == 0) {
        return NULL;
    }
    else if (size == 0) {
        return NULL;
    }

    for (int cnt = 0; cnt < rank; ++cnt)
    {
        total_size *= dims[cnt];
    }

    total_size *= size;

    return malloc(total_size);
}

/** release memory.
 *
 * @param   buf         [void*][in] address of allocated memory.
 * @return  result status.
 */
int SGTK_free(void *buf)
{
    if (buf == NULL) {
        return ERROR_HDF_FILE_CLOSE;
    }
    free(buf);
    buf = NULL;

    return SUCCESS;
}

/** get latidute and longitude.
 *
 * @param   HDF_file_id     [hid_t][in]     hdf5 file id.
 * @param   latlon          [float**][out]  latitude and longitude value.
 * @return  number of elements (number of latitude * number of longitude).
 */
int SGTK_get_global_latlon(hid_t HDF_file_id, float *latlon)
{
    int result = SUCCESS;

    // get granule id.
    char granule_id[1024];
    memset(granule_id, 0, sizeof(granule_id));
    result = SGTK_get_meta_string(HDF_file_id, (char*)"/Global_attributes", (char*)"Product_file_name", (char*)granule_id);
    if(result!=SUCCESS) {
        return ERROR_PARAM;
    }

    int ireso = 0;
    int ih = 0;
    int iv = 0;
    char reso = ' ';
    char h[4], v[4];
    memset(h, 0, sizeof(h));
    memset(v, 0, sizeof(v));
    bool isTile = true;
    switch(granule_id[20]) {
    case 'T':
        // eqa tile.
        strncpy(h, &granule_id[23], 2);
        strncpy(v, &granule_id[21], 2);
        ih = atoi(h);
        iv = atoi(v);
        reso = granule_id[35];
        switch(reso) {
        case 'K':   // 1km
            ireso=0;
            break;
        case 'H':   // 500m
            ireso=1;
            break;
        case 'Q':   // 250m
            ireso=2;
            break;
        default:
            return ERROR_PARAM;
            break;
        }
        calcLatLon_eqa_tile(ireso, ih, iv, latlon);
        break;
    case 'A':
    case 'X':
        // eqa.
        strncpy(h, &granule_id[23], 2);
        strncpy(v, &granule_id[21], 2);
        ih = atoi(h);
        iv = atoi(v);
        reso = granule_id[35];
        switch(reso) {
        case 'K':   // 1km
            isTile = true;
            ireso=0;
            break;
        case 'H':   // 500m
            isTile = true;
            ireso=1;
            break;
        case 'Q':   // 250m
            isTile = true;
            ireso=2;
            break;
        case 'C':   // 1/12
            isTile = false;
            ireso=0;
            break;
        case 'F':   // 1/24
            isTile = false;
            ireso=1;
            break;
        default:
            return ERROR_PARAM;
            break;
        }
        if(isTile==true) {
            calcLatLon_eqa_tile(ireso, ih, iv, latlon);
        } else {
            calcLatLon_eqa_binning(ireso, latlon);
        }
        break;
    case 'D':
        // eqr.
        reso = granule_id[35];
        switch(reso) {
        case 'C':   // 1/12
            ireso=0;
            break;
        case 'F':   // 1/24
            ireso=1;
            break;
        default:
            return ERROR_PARAM;
            break;
        }
        calcLatLon_eqr(ireso, latlon);
        break;
    case 'N':
    case 'S':
        // polar stereo.
        reso = granule_id[35];
        switch(reso) {
        case 'C':   // 1/12
            ireso=0;
            break;
        case 'F':   // 1/24
            ireso=1;
            break;
        default:
            return ERROR_PARAM;
            break;
        }
        calcLatLon_ps(ireso, latlon);
        break;
    default:
        result = ERROR_PARAM;
        break;
    }

    return result;
}

/** load configuration file.
 *
 * @param   config_file     [char*][in]     configuration file name.
 * @return  result status.
 */
int SGTK_load_config(char *file_name)
{
    return load_config(file_name);
}

/** get dimensions.
 *
 * @param   HDF_file_id     [int][in]       hdf5 file id.
 * @param   label           [char*][in]     Dataset Access Label.
 * @param   num             [int**][out]    elements of each dimensions.
 * @return  number of dimensions.
 */
int SGTK_get_dims(hid_t HDF_file_id, char *access_label, int **num)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    // get_ds_dims()
    int                     rank = 0;
    std::vector<hsize_t>    dims;

    int ret = HDF::get_ds_dims(HDF_file_id, label, &rank, &dims);
    if (ret < 0)
    {
        return ret;
    }

    *num = (int*)malloc(rank * sizeof(int));
    if (*num == NULL)
    {
        return ERROR_MEMSIZE;
    }

    int     *val = *num;
    for (int cnt = 0; cnt < rank; ++cnt)
    {
        val[cnt] = dims.at(cnt);
    }

    return rank;
}

/** get signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]       hdf5 file id.
 * @param   label           [char*][in]     Dataset Access Label.
 * @param   get_value       [int**][out]    address of dataset values.
 * @return  result status.
 */
int SGTK_get_dataset_int(hid_t HDF_file_id, char *access_label, int **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<int> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return ret;
    }
    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(int);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    int vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    int *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 8-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int8_t**][out]     address of dataset values.
 * @return  result status.
 */
int SGTK_get_dataset_int8(hid_t HDF_file_id, char *access_label, int8_t **get_value)
{
	
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<int8_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(int8_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    int8_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int8_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int8_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    int8_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 16-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int16_t**][out]    address of dataset values.
 * @return  result status.
 */
int SGTK_get_dataset_int16(hid_t HDF_file_id, char *access_label, int16_t **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<int16_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(int16_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    int16_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int16_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int16_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    int16_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 32-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int32_t**][out]    address of dataset values.
 * @return  result status.
 */
int SGTK_get_dataset_int32(hid_t HDF_file_id, char *access_label, int32_t **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<int32_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(int32_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    int32_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int32_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int32_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    int32_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int**][out]        address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_uint(hid_t HDF_file_id, char *access_label, uint **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<uint> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(uint);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    uint vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    uint *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 8-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int8_t**][out]     address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_uint8(hid_t HDF_file_id, char *access_label, uint8_t **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<uint8_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(uint8_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    uint8_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint8_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint8_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    uint8_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 16-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int16_t**][out]    address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_uint16(hid_t HDF_file_id, char *access_label, uint16_t **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<uint16_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(uint16_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    uint16_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint16_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint16_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    uint16_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get 32-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   get_value       [int32_t**][out]    address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_uint32(hid_t HDF_file_id, char *access_label, uint32_t **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    //call get_dataset_numeric_value
    std::vector<hsize_t> vdims;
    std::vector<uint32_t> value;

    int ret = HDF::get_dataset_numeric_value(HDF_file_id, label, vdims, value);
    if (ret < SUCCESS) {
        return (int)ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < vdims.size(); ++cnt) size *= vdims.at(cnt);
    size *= sizeof(uint32_t);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }

    // Get Min value
    uint32_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint32_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint32_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    uint32_t *val = *get_value;
    for (size_t cnt = 0; cnt < value.size(); ++cnt)
    {
        val[cnt] = value.at(cnt);
        if (range_flag == true){
            if (value[cnt] < vmin || vmax < value[cnt]){
                value[cnt] = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            val[cnt] = reverse_endian(value.at(cnt));
        }
    }

    return SUCCESS;
}

/** get float dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   get_value       [float**][out]      address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_float(hid_t HDF_file_id, char *access_label, int kind, float **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind != SGTK_KIND_RADIANCE) && (kind != SGTK_KIND_REFLECTANCE) && (kind != SGTK_KIND_OTHER) )
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>    dims;
    std::vector<float>      value;

    int ret = HDF::get_float_value_with_slope_and_offset(HDF_file_id,
                                                         label,
                                                         dims,
                                                         value,
                                                         kind);
    if (ret != SUCCESS)
    {
        return ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < dims.size(); ++cnt)  size *= dims.at(cnt);
    size *= sizeof(float);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }
    float   *val = *get_value;
    vec2ptr(value, val, value.size(), byte_order_obj.get_flg(HDF_file_id));

    return SUCCESS;
}

/** get 32-bit float dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   get_value       [float**][out]      address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_float32(hid_t HDF_file_id, char *label, int kind, float **get_value)
{
	
    return SGTK_get_dataset_float(HDF_file_id, label, kind, get_value);
}

/** get 64-bit float dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   get_value       [double**][out]     address of dataset values.
 * @return  result status.
 * */
int SGTK_get_dataset_float64(hid_t HDF_file_id, char *access_label, int kind, double **get_value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind != SGTK_KIND_RADIANCE) && (kind != SGTK_KIND_REFLECTANCE) && (kind != SGTK_KIND_OTHER) )
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>    dims;
    std::vector<double>     value;

    int ret = HDF::get_double_value_with_slope_and_offset(HDF_file_id,
                                                          label,
                                                          dims,
                                                          value,
                                                          kind);
    if (ret != SUCCESS)
    {
        return ret;
    }

    size_t      size = 1;
    for (size_t cnt = 0; cnt < dims.size(); ++cnt)  size *= dims.at(cnt);
    size *= sizeof(double);
    if (*get_value == NULL)
    {
        return ERROR_MEMSIZE;
    }
    double  *val = *get_value;
    vec2ptr(value, val, value.size(), byte_order_obj.get_flg(HDF_file_id));

    return SUCCESS;
}

/** get latitude and longitude.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label_lat       [char*][in]         Dataset Access Label of latitude.
 * @param   label_lon       [char*][in]         Dataset Access Label of longitude.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   start_no        [int][in]           start scan number.
 * @param   end_no          [int][in]           end scan number.
 * @param   latlon          [float**][out]      address of dataset values.
 * @return  number of elements (number of latitude * number of longitude).
 * */
int SGTK_get_dataset_latlon(hid_t HDF_file_id, char *access_label_lat, char *access_label_lon, int kind, int start_no, int end_no, float **latlon)
{
	std::string label_lat;
	std::string label_lon;
	
	SGKT_get_label_to_path(access_label_lat, label_lat);
	SGKT_get_label_to_path(access_label_lon, label_lon);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label_lat.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (label_lon.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind < SGTK_KIND_RADIANCE) || (kind > SGTK_KIND_REFLECTANCE))
    {
        return ERROR_PARAM;
    }
    if ((start_no < 0) || (end_no < 0))
    {
        return ERROR_PARAM;
    }
    if ((start_no == end_no) || (start_no > end_no))
    {
        return ERROR_PARAM;
    }
    

    // latitude get
    std::vector<hsize_t>    lat_dims;
    std::vector<float>      lat_value;
    std::vector<int>        lat_interval;

    int ret = HDF::get_lat_lon_with_interval(HDF_file_id,
                                             label_lat,
                                             lat_interval,
                                             kind, 
                                             lat_dims,
                                             lat_value);
    if (ret != SUCCESS)
    {
        return ret;
    }

    // longitude get
    std::vector<hsize_t>    lon_dims;
    std::vector<float>      lon_value;
    std::vector<int>        lon_interval;

    ret = HDF::get_lat_lon_with_interval(HDF_file_id,
                                         label_lon,
                                         lon_interval,
                                         kind,
                                         lon_dims,
                                         lon_value);
    if (ret != SUCCESS)
    {
        return ret;
    }

    // range check
    if (lat_dims.size() != lon_dims.size())
    {
        return ERROR_GET_DS_DIMS;
    }
    std::vector<hsize_t>    dims(3,1);
    if( (lat_dims.size()==2) )
    {
        dims.at(1) = lat_dims.at(0);    // lines
        dims.at(2) = lat_dims.at(1);    // pixels
    } else
    if( (lat_dims.size()==3) )
    {
        dims.at(0) = lat_dims.at(0);    // lenses
        dims.at(1) = lat_dims.at(1);    // lines
        dims.at(2) = lat_dims.at(2);    // pixels
    } else {
        return ERROR_RANGE_DIMS;
    }
    if (*latlon == NULL)
    {
        return ERROR_MEMSIZE;
    }

	std::string label;
	SGKT_get_label_to_path("GC1_L1A_NP_ATR_PRODUCT_FILE_NAME", label);
	
    // get granule id.
    char granule_id[1024];
    memset(granule_id, 0, sizeof(granule_id));
    ret = SGTK_get_meta_string(HDF_file_id, (char*)"GC1_L1A_NP_GRP_GLOBAL_ATTRIBUTES", (char*)"GC1_L1A_NP_ATR_PRODUCT_FILE_NAME", (char*)granule_id);
    if(ret!=SUCCESS) {
        return ERROR_PARAM;
    }
    // get lines par scan
    int lines = get_lines_par_scan((char*)&granule_id, label_lat);
    if(lines<0) {
        return lines;
    }

    LatLonStrct     *lat_lon = (LatLonStrct *)*latlon;

    // get lat lon data set value
    size_t setlines  = dims.at(1);
    size_t setpixels = dims.at(2);
    for(size_t lens=0; lens<(size_t)dims.at(0); ++lens) {
        for(size_t line=start_no*lines; line<(size_t)((end_no)*lines); ++line) {
            for(size_t pixel=0; pixel<(size_t)dims.at(2); ++pixel) {
                if(byte_order_obj.get_flg(HDF_file_id)==true) {
                    lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lat = lat_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel);
                    lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lon = lon_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel);
                } else {
                    lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lat = reverse_endian(lat_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel));
                    lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lon = reverse_endian(lon_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel));
                }
            }
        }
    }

    return SUCCESS;
}

/** get attribute string.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [char*][out]        attribute value.
 * @return  result status.
 */
int SGTK_get_meta_string(hid_t HDF_file_id, char *access_label, char *access_meta, char *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<std::string>    val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_string_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_string_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    strcpy(value, val.at(0).c_str());

    return SUCCESS;
}

/** get signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int*][out]         attribute value.
 * @return  result status.
 */
int SGTK_get_meta_int(hid_t HDF_file_id, char *access_label, char *access_meta, int *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<int>            val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 8-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int8_t*][out]      attribute value.
 * @return  result status.
 */
int SGTK_get_meta_int8(hid_t HDF_file_id, char *access_label, char *access_meta, int8_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<int8_t>         val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 16-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int16_t*][out]     attribute value.
 * @return  result status.
 */
int SGTK_get_meta_int16(hid_t HDF_file_id, char *access_label, char *access_meta, int16_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<int16_t>        val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 32-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int32_t*][out]     attribute value.
 * @return  result status.
 */
int SGTK_get_meta_int32(hid_t HDF_file_id, char *access_label, char *access_meta, int32_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<int32_t>        val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint*][out]        attribute value.
 * @return  result status.
 */
int SGTK_get_meta_uint(hid_t HDF_file_id, char *access_label, char *access_meta, uint *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<uint>           val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 8-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint8_t*][out]     attribute value.
 * @return  result status.
 */
int SGTK_get_meta_uint8(hid_t HDF_file_id, char *access_label, char *access_meta, uint8_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<uint8_t>        val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 16-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint16_t*][out]    attribute value.
 * @return  result status.
 */
int SGTK_get_meta_uint16(hid_t HDF_file_id, char *access_label, char *access_meta, uint16_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<uint16_t>       val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 32-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint32_t*][out]    attribute value.
 * @return  result status.
 */
int SGTK_get_meta_uint32(hid_t HDF_file_id, char *access_label, char *access_meta, uint32_t *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }

    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<uint32_t>       val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get float attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [float*][out]       attribute value.
 * @return  result status.
 */
int SGTK_get_meta_float(hid_t HDF_file_id, char *access_label, char *access_meta, float *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<float>          val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get 32-bit float attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [float*][out]       attribute value.
 * @return  result status.
 */
int SGTK_get_meta_float32(hid_t HDF_file_id, char *access_label, char *access_meta, float *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<float>          val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** get double attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [double*][out]      attribute value.
 * @return  result status.
 */
int SGTK_get_meta_float64(hid_t HDF_file_id, char *access_label, char *access_meta, double *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>        dims;
    std::vector<double>         val;
    int                         ret = SUCCESS;

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::get_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::get_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        *value = val.at(0);
    } else {
        *value = reverse_endian(val.at(0));
    }

    return SUCCESS;
}

/** create initial dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   keylabel        [char*][in]         Dataset Access Label.
 * @param   dims            [int**][in]         elements of each dimensions.
 * @param   rank            [int*][in]          number of dimensions.
 * @param   dat_type        [int][in]           HDF5 data type.
 * @param   dim_lbl         [char**][in]        dimension label of each dimension.
 * @return  result status.
 */
int SGTK_set_dims(int  HDF_file_id,
                  char *access_label,
                  int  *dims, 
                  int  rank, 
                  int  dat_type,
                  char **dim_lbl)
{
    int             ret = SUCCESS;
	std::string     label;
	
	SGKT_get_label_to_path(access_label, label);

    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }
    if (rank <= 0)
    {
        return ERROR_PARAM;
    }
	
    if ((dat_type < SGTK_H5T_STD_I8_TYPE) || (dat_type > SGTK_H5T_C_S1_TYPE))
    {
        return ERROR_PARAM;
    }

    PrdDatasetInfo  dts; 

    switch (rank)
    {
    case 1:
        if (dim_lbl[0] == NULL){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        break;
    case 2:
        if ((dim_lbl[0] == NULL) || (dim_lbl[1] == NULL)){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        dts.label.push_back(dim_lbl[1]);
        break;
    case 3:
        if ((dim_lbl[0] == NULL) || (dim_lbl[1] == NULL) || (dim_lbl[2] == NULL)){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        dts.label.push_back(dim_lbl[1]);
        dts.label.push_back(dim_lbl[2]);
        break;
    case 4:
        if ((dim_lbl[0] == NULL) || (dim_lbl[1] == NULL) ||
            (dim_lbl[2] == NULL) || (dim_lbl[3] == NULL)){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        dts.label.push_back(dim_lbl[1]);
        dts.label.push_back(dim_lbl[2]);
        dts.label.push_back(dim_lbl[3]);
        break;
    case 5:
        if ((dim_lbl[0] == NULL) || (dim_lbl[1] == NULL) ||
            (dim_lbl[2] == NULL) || (dim_lbl[3] == NULL) || (dim_lbl[4] == NULL)){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        dts.label.push_back(dim_lbl[1]);
        dts.label.push_back(dim_lbl[2]);
        dts.label.push_back(dim_lbl[3]);
        dts.label.push_back(dim_lbl[4]);
        break;
    case 6:
        if ((dim_lbl[0] == NULL) || (dim_lbl[1] == NULL) ||
            (dim_lbl[2] == NULL) || (dim_lbl[3] == NULL) ||
            (dim_lbl[4] == NULL) || (dim_lbl[5] == NULL)){
            ret = ERROR_PARAM;
            break;
        }
        dts.label.push_back(dim_lbl[0]);
        dts.label.push_back(dim_lbl[1]);
        dts.label.push_back(dim_lbl[2]);
        dts.label.push_back(dim_lbl[3]);
        dts.label.push_back(dim_lbl[4]);
        dts.label.push_back(dim_lbl[5]);
        break;
    }
    if (ret != SUCCESS)
    {
        return ret;
    }

    int8_t          i8_type  = 0;
    uint8_t         u8_type  = 0;
    int16_t         i16_type = 0;
    uint16_t        u16_type = 0;
    int32_t         i32_type = 0;
    uint32_t        u32_type = 0;
    int64_t         i64_type = 0;
    uint64_t        u64_type = 0;
    float           f32_type = 0;
    double          f64_type = 0;
    std::string     str_type;

    hsize_t type_size=0;
    dts.path = label;
    switch (dat_type)
    {
    case SGTK_H5T_STD_I8_TYPE:
        dts.type = HDF::get_typeid(i8_type);
        type_size = sizeof(char);
        break;
    case SGTK_H5T_STD_U8_TYPE:
        dts.type = HDF::get_typeid(u8_type);
        type_size = sizeof(unsigned char);
        break;
    case SGTK_H5T_STD_I16_TYPE:
        dts.type = HDF::get_typeid(i16_type);
        type_size = sizeof(short);
        break;
    case SGTK_H5T_STD_U16_TYPE:
        dts.type = HDF::get_typeid(u16_type);
        type_size = sizeof(unsigned short);
        break;
    case SGTK_H5T_STD_I32_TYPE:
        dts.type = HDF::get_typeid(i32_type);
        type_size = sizeof(int);
        break;
    case SGTK_H5T_STD_U32_TYPE:
        dts.type = HDF::get_typeid(u32_type);
        type_size = sizeof(unsigned int);
        break;
    case SGTK_H5T_STD_I64_TYPE:
        dts.type = HDF::get_typeid(i64_type);
        type_size = sizeof(long);
        break;
    case SGTK_H5T_STD_U64_TYPE:
        dts.type = HDF::get_typeid(u64_type);
        type_size = sizeof(unsigned long);
        break;
    case SGTK_H5T_IEEE_F32_TYPE:
        dts.type = HDF::get_typeid(f32_type);
        type_size = sizeof(float);
        break;
    case SGTK_H5T_IEEE_F64_TYPE:
        dts.type = HDF::get_typeid(f64_type);
        type_size = sizeof(double);
        break;
    default:
        return ERROR_PARAM;
        break;
    }

    dts.rank = rank;

    // Set dimension size.
    size_t      size = 1;
    for (int cnt = 0; cnt < dts.rank; ++cnt)
    {
        dts.dims.push_back(dims[cnt]);
        size *= dims[cnt];
    }

    // Set chunk size
    for( int cnt = 0; cnt < 6; cnt++) {
        dts.cdims.push_back(conf_obj.chunk);
    }

    // Set compression level.
    dts.level = (CompLevel)conf_obj.compress_level;

    // Set initial value.
    dts.p_value = malloc(size*type_size);

    HDF::set_dataset(HDF_file_id, dts);

    free(dts.p_value);

    return SUCCESS;
}

/** set integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [int*][in]          dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_int(hid_t HDF_file_id, char *access_label, int *dims, int rank, int *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    int vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<int> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }

    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 8-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [int8_t*][in]       dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_int8(hid_t HDF_file_id, char *access_label, int * dims, int rank, int8_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    int vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<int8_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }

    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 16-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [int16_t*][in]      dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_int16(hid_t HDF_file_id, char *access_label, int * dims, int rank, int16_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    int16_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int16_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int16_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<int16_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 32-bit signed integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [int32_t*][in]      dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_int32(hid_t HDF_file_id, char *access_label, int * dims, int rank, int32_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    int32_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    int32_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    int32_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<int32_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [uint*][in]         dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_uint(hid_t HDF_file_id, char *access_label, int * dims, int rank, uint *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    uint vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<uint> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 8-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [uint8_t*][in]      dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_uint8(hid_t HDF_file_id, char *access_label, int * dims, int rank, uint8_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    uint8_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint8_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint8_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<uint8_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 16-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [uint16_t*][in]     dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_uint16(hid_t HDF_file_id, char *access_label, int * dims, int rank, uint16_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    uint16_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint16_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint16_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<uint16_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set 32-bit unsigned integer dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [int32_t*][in]      dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_uint32(hid_t HDF_file_id, char *access_label, int *dims, int rank, uint32_t *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }

    // set in_param
    std::vector<hsize_t> vdims;
    std::vector<int> chunks;
    hsize_t size = 1;

    CompLevel cmplvl = (CompLevel)conf_obj.compress_level;

    for ( int i = 0; i < rank; i++) {
            size *= dims[i];
            vdims.push_back(dims[i]);
            chunks.push_back(conf_obj.chunk);
    }

    // Get Min value
    uint32_t vmin = 0;
    bool minret = HDF::get_min_attr_value(HDF_file_id, label, vmin);
    // Get Max value
    uint32_t vmax = 0;
    bool maxret = HDF::get_max_attr_value(HDF_file_id, label, vmax);
    // Get error value.
    uint32_t err = 0;
    bool errret = HDF::get_error_value(HDF_file_id, label, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    std::vector<uint32_t> vvalue(size);
    for ( unsigned int i = 0; i < size; i++) { 
        vvalue.at(i) = *(value+i);
        if (range_flag == true){
            if (vvalue.at(i) < vmin || vmax < vvalue.at(i)){
                vvalue.at(i) = err;
            }
        }
        if(byte_order_obj.get_flg(HDF_file_id)==false) {
            vvalue.at(i) = reverse_endian(*(value+i));
        }
    }

    // call set_dataset_numeric_value
    int ret = HDF::set_dataset_numeric_value(HDF_file_id, label, vdims, vvalue, cmplvl, chunks);
    return (int)ret;
}

/** set float dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [float*][in]        dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_float(hid_t HDF_file_id, char *access_label, int kind, int *dims, int rank, float *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind != SGTK_KIND_RADIANCE) && (kind != SGTK_KIND_REFLECTANCE) && (kind != SGTK_KIND_OTHER) )
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }
    if (rank <= 0)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>    dim;
    CompLevel               cmplvl = (CompLevel)conf_obj.compress_level;
    int                     chunks = conf_obj.chunk;
    std::vector<int>        vchunks;
    hsize_t                 size = 1;

    for ( int i = 0; i < rank; i++) {
        size *= dims[i];
        dim.push_back(dims[i]);
        vchunks.push_back(conf_obj.chunk);
    }
    std::vector<float>      val(size);

    int ret = SUCCESS;
    if(kind==SGTK_KIND_OTHER) {
        ptr2vec(value, val, size, byte_order_obj.get_flg(HDF_file_id));
        ret = HDF::set_dataset_numeric_value(HDF_file_id, label, dim, val, cmplvl, vchunks);
    } else {
        for (  unsigned int i = 0; i < size; i++) { 
            val.at(i) = value[i];
        }

        ret = HDF::set_float_value_without_slope_and_offset(HDF_file_id, label, dim, val, kind, cmplvl, chunks);
    }

    return ret;
}

/** set 32-bit float dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [flaot*][in]        dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_float32(hid_t HDF_file_id, char *label, int kind, int *dims, int rank, float *value)
{
    return SGTK_set_dataset_float(HDF_file_id, label, kind, dims, rank, value);
}

/** set double dataset.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [double*][in]       dataset value.
 * @return  result status.
 */
int SGTK_set_dataset_float64(hid_t HDF_file_id, char *access_label, int kind, int *dims, int rank, double *value)
{
	std::string label;
	
	SGKT_get_label_to_path(access_label, label);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind != SGTK_KIND_RADIANCE) && (kind != SGTK_KIND_REFLECTANCE) && (kind != SGTK_KIND_OTHER) )
    {
        return ERROR_PARAM;
    }
    if (dims == NULL)
    {
        return ERROR_PARAM;
    }
    if (rank <= 0)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    std::vector<hsize_t>    dim;
    CompLevel               cmplvl = (CompLevel)conf_obj.compress_level;
    int                     chunks = conf_obj.chunk;
    std::vector<int>        vchunks;
    hsize_t                 size = 1;

    for ( int i = 0; i < rank; i++) {
        size *= dims[i];
        dim.push_back(dims[i]);
        vchunks.push_back(conf_obj.chunk);
    }
    std::vector<double>     val(size);

    int ret = SUCCESS;
    if(kind==SGTK_KIND_OTHER) {
        ptr2vec(value, val, size, byte_order_obj.get_flg(HDF_file_id));
        ret = HDF::set_dataset_numeric_value(HDF_file_id, label, dim, val, cmplvl, vchunks);
    } else {
        for (  unsigned int i = 0; i < size; i++) { 
            val.at(i) = value[i];
        }
        ret = HDF::set_double_value_without_slope_and_offset(HDF_file_id, label, dim, val, kind, cmplvl, chunks);
    }


    return ret;
}

/** set latitude and longitude.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label_lat       [char*][in]         Dataset Access Label of latitude.
 * @param   label_lon       [char*][in]         Dataset Access Label of longitude.
 * @param   kind            [int][in]           kind of value (radiance/reflectance/other).
 * @param   start_no        [int][in]           start scan number.
 * @param   end_no          [int][in]           end scan number.
 * @param   set_num         [int][in]           number of elements (number of latitude * number of longitude).
 * @param   latlon          [float*][in]        address of dataset values.
 * @return  result status.
 * */
int SGTK_set_dataset_latlon(hid_t HDF_file_id, char *access_label_lat, char *access_label_lon, int kind, int start_no, int end_no, int set_num, float *latlon)
{
	std::string label_lat;
	std::string label_lon;
	
	SGKT_get_label_to_path(access_label_lat, label_lat);
	SGKT_get_label_to_path(access_label_lon, label_lon);
	
    // parameter check
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label_lat.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (label_lon.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if ((kind < SGTK_KIND_RADIANCE) || (kind > SGTK_KIND_REFLECTANCE))
    {
        return ERROR_PARAM;
    }
    if ((start_no < 0) || (end_no < 0))
    {
        return ERROR_PARAM;
    }
    if ((start_no == end_no) || (start_no > end_no))
    {
        return ERROR_PARAM;
    }
    if (set_num == 0)
    {
        return ERROR_PARAM;
    }
    if (((end_no - start_no) + 1) == (set_num / 2))
    {
        return ERROR_PARAM;
    }

    if (latlon == NULL)
    {
        return ERROR_PARAM;
    }

    // latitude get
    std::vector<hsize_t>    lat_dims;
    std::vector<float>      lat_value;
    std::vector<int>        lat_interval;

    int ret = HDF::get_lat_lon_with_interval(HDF_file_id,
                                             label_lat,
                                             lat_interval,
                                             kind,
                                             lat_dims,
                                             lat_value);
    if (ret != SUCCESS)
    {
        return ret;
    }

    // longitude get
    std::vector<hsize_t>    lon_dims;
    std::vector<float>      lon_value;
    std::vector<int>        lon_interval;

    ret = HDF::get_lat_lon_with_interval(HDF_file_id,
                                         label_lon,
                                         lon_interval,
                                         kind,
                                         lon_dims,
                                         lon_value);
    if (ret != SUCCESS)
    {
        return ret;
    }
    // range check
    if (lat_dims.size() != lon_dims.size())
    {
        return ERROR_GET_DS_DIMS;
    }
    
    std::vector<hsize_t>    dims(3,1);
    if( (lat_dims.size()==2) )
    {
        dims.at(1) = lat_dims.at(0);    // lines
        dims.at(2) = lat_dims.at(1);    // pixels
    } else
    if( (lat_dims.size()==3) )
    {
        dims.at(0) = lat_dims.at(0);    // lenses
        dims.at(1) = lat_dims.at(1);    // lines
        dims.at(2) = lat_dims.at(2);    // pixels
    } else {
        return ERROR_RANGE_DIMS;
    }

    // get granule id.
    char granule_id[1024];
    memset(granule_id, 0, sizeof(granule_id));
    ret = SGTK_get_meta_string(HDF_file_id, (char*)"GC1_L1A_NP_GRP_GLOBAL_ATTRIBUTES", (char*)"GC1_L1A_NP_ATR_PRODUCT_FILE_NAME", (char*)granule_id);
    if(ret!=SUCCESS) {
        return ERROR_PARAM;
    }
    // get lines par scan
    int lines = get_lines_par_scan((char*)&granule_id, label_lat);
    if(lines<0) {
        return lines;
    }

    LatLonStrct     *lat_lon = (LatLonStrct *)latlon;

    // set lat lon data set value
    size_t setlines  = dims.at(1);
    size_t setpixels = dims.at(2);
    for(size_t lens=0; lens<(size_t)dims.at(0); ++lens) {
        for(size_t line=start_no*lines; line<(size_t)((end_no)*lines); ++line) {
            for(size_t pixel=0; pixel<(size_t)dims.at(2); ++pixel) {
                if(byte_order_obj.get_flg(HDF_file_id)==true) {
                    lat_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel) = lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lat;
                    lon_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel) = lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lon;
                } else {
                    lat_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel) = reverse_endian(lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lat);
                    lon_value.at((lens*setlines*setpixels)+(line*setpixels)+pixel) = reverse_endian(lat_lon[(lens*setlines*setpixels)+(line*setpixels)+pixel].lon);
                }
            }
        }
    }

    // lat set
    ret = HDF::set_lat_lon_without_interval(HDF_file_id, label_lat, lat_dims, lat_value, kind);
    if (ret != SUCCESS)
    {
        return ret;
    }

    // lon set
    ret = HDF::set_lat_lon_without_interval(HDF_file_id, label_lon, lon_dims, lon_value, kind);
    if (ret != SUCCESS)
    {
        return ret;
    }

    return SUCCESS;
}

/** set string attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [char*][in]         attribute value.
 * @return  result status.
 */
int SGTK_set_meta_string(hid_t HDF_file_id, char *access_label, char *access_meta, char *value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (value == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<std::string>    val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    val.push_back(value);

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_string_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_string_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int*][in]          attribute value.
 * @return  result status.
 */
int SGTK_set_meta_int(hid_t HDF_file_id, char *access_label, char *access_meta, int value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<int>            val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 8-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int8_t*][in]       attribute value.
 * @return  result status.
 */
int SGTK_set_meta_int8(hid_t HDF_file_id, char *access_label, char *access_meta, int8_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<int8_t>         val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 16-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int16_t*][in]      attribute value.
 * @return  result status.
 */
int SGTK_set_meta_int16(hid_t HDF_file_id, char *access_label, char *access_meta, int16_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<int16_t>        val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 32-bit signed integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [int32_t*][in]      attribute value.
 * @return  result status.
 */
int SGTK_set_meta_int32(hid_t HDF_file_id, char *access_label, char *access_meta, int32_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<int32_t>        val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   dims            [int*][in]          elements of each dimensions.
 * @param   rank            [int][in]           number of dimensions.
 * @param   value           [uint*][in]         attribute value.
 * @return  result status.
 */
int SGTK_set_meta_uint(hid_t HDF_file_id, char *access_label, char *access_meta, uint value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<uint>           val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 8-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint8_t*][in]      attribute value.
 * @return  result status.
 */
int SGTK_set_meta_uint8(hid_t HDF_file_id, char *access_label, char *access_meta, uint8_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<uint8_t>        val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 16-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint16_t*][in]     attribute value.
 * @return  result status.
 */
int SGTK_set_meta_uint16(hid_t HDF_file_id, char *access_label, char *access_meta, uint16_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<uint16_t>       val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 32-bit unsigned integer attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [uint32_t*][in]     attribute value.
 * @return  result status.
 */
int SGTK_set_meta_uint32(hid_t HDF_file_id, char *access_label, char *access_meta, uint32_t value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<uint32_t>       val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set float attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [float*][in]        attribute value.
 * @return  result status.
 */
int SGTK_set_meta_float(hid_t HDF_file_id, char *access_label, char *access_meta, float value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
    
	if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<float>          val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set 32-bit float attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [float*][in]        attribute value.
 * @return  result status.
 */
int SGTK_set_meta_float32(hid_t HDF_file_id, char *access_label, char *access_meta, float value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<float>          val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** set double attribute.
 *
 * @param   HDF_file_id     [int][in]           hdf5 file id.
 * @param   label           [char*][in]         Dataset Access Label.
 * @param   meta            [char*][in]         attribute name.
 * @param   value           [double*][in]       attribute value.
 * @return  result status.
 */
int SGTK_set_meta_float64(hid_t HDF_file_id, char *access_label, char *access_meta, double value)
{
	std::string     label;
	std::string     meta;
	
	SGKT_get_label_to_path(access_label, label);
	SGKT_get_label_to_path(access_meta, meta);
	
    if (HDF_file_id < 0)
    {
        return ERROR_PARAM;
    }
    if (label.c_str() == NULL)
    {
        return ERROR_PARAM;
    }
    if (meta.c_str() == NULL)
    {
        return ERROR_PARAM;
    }

    int                         ret = SUCCESS;
    std::vector<double>         val;
    std::vector<hsize_t>        dims;

    dims.push_back(1);
    if(byte_order_obj.get_flg(HDF_file_id)==true) {
        val.push_back(value);
    } else {
        val.push_back(reverse_endian(value));
    }

    if (true == HDF::exist_group(HDF_file_id, label))
    {
        ret = HDF::set_group_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }
    else
    {
        ret = HDF::set_dataset_attr_numeric_value(HDF_file_id, label, meta, dims, val);
    }

    return ret;
}

/** make parameter table
 *
 * @param   parameter_file  [char*][in]  parameter file name
 * @return  result status.
 */
int SGTK_make_parameter_table( const std::string parameter_file )
{
	FILE* fp = NULL;
	char buf[1024];
	char* key = NULL;
	char* value = NULL;
	std::string readLabel;
	std::string readPath;
	
	memset(buf, 0x00, sizeof(buf));
	
	// t@CI[vip[^t@CŒj
	fp = fopen(parameter_file.c_str(), "r");
	if(fp == NULL){
		return -1;
	}
	
	//xe[uɓo^
	while( fgets(buf, sizeof(buf), fp) != NULL){
		if(strstr(buf, "=") != NULL){
			key = strtok(buf, "=");
			value = strtok(NULL, "=");

			readLabel = std::string( strtok(key, "\r\n") );
			readPath  = std::string( strtok(value, "\r\n") );
			
			//printf("key=[%s]\n", readLabel.c_str());
			//printf("value=[%s]\n", readPath.c_str());
			
			// O[oϐ̃e[uɓo^
			SGLI_ACCESS_LABEL_TABLE.insert(std::pair<std::string, std::string>(readLabel, readPath));
		}
		
		memset(buf, 0x00, sizeof(buf));
		key = NULL;
		value = NULL;
	}
	
	return 0;
}

/** get label to path.
 *
 * @param   label           [char*][in]               Dataset Access Label.
 * @param   path            [std::string][out]        Dataset Path.
 * @return  result status.
 */
int SGKT_get_label_to_path(const std::string label, std::string& path)
{
	int ret = -1;
	std::map<std::string, std::string>::iterator itr;
	
	itr = SGLI_ACCESS_LABEL_TABLE.find(label);
	
	//f[^Zbg̃xꍇApXݒ肷
	if(itr != SGLI_ACCESS_LABEL_TABLE.end()){
		path = itr->second;
		ret = SUCCESS;
	}
	
	return ret;
}
