/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK_date_time.cpp                                        */
/*  Abstract     : I/O toolkit date and time  function                       */
/*  Function List                                                            */
/*      com_get_leapsecond_formal() : Load leap second file                  */
/*      com_time_tai93_utc()        : Convert from TAI93 to UTC              */
/*      com_time_utc_tai93()        : Convert from UTC to TAI93              */
/*      get_jd_and_mjd()            : Get Julian day and Modified Julian day */
/*      get_tai93_begintime()       : Get TAI93 standard value               */
/*      get_leapsecond_utc_tai()    : Get leap second                        */
/*      com_release_leapsecond()    : Release leap second table              */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include <float.h>

#include "SGTK_date_time.h"
#include "SGTK_ret_value.h"


/* Unit Test */
#ifdef UNIT_TEST
#include <assert.h>
#define DPRINTF(...) \
    printf("%s(%d)[%s]:", __FILE__, __LINE__, __func__); \
    printf(__VA_ARGS__)
#else
#define DPRINTF(fmt, ...)
#endif

/** size of leapsecond table. */
static int g_leap_second_table_size = 0;

/** leapsecond table. */
static double (*p_leap_second_table)[LEAPSECOND_COL_NUM] = NULL;


/** Get TAI93 standard value.
 * (Internal function)
 * @param p_tai_begintime     [out]TAI93 standard value
 * @param p_tai_begin_leapsec [out]Leap Second of TAI93 standard value
 */
static void get_tai93_begintime(
    struct tm *p_tai_begintime,
    double *p_tai_begin_leapsec);

/** Get Leapsecond Formal Ver.
 * Leapsecond is acquired and it maintains it in the memory.
 * 
 * @param  leap_second_file_name [in]FILE NAME OF LEAPSECOND FILE.
 * @return NORMAL
 * @return FOPEN_ERROR_LEAPSECOND
 * @return MEM_ERROR_LEAPSECOND
 * @return NO_EXIST_ERROR_LEAPSECOND_TABLE
 */
int com_get_leapsecond_formal(const char *leap_second_file_name)
{
    FILE *p_file = NULL;
    char buffer[256] = "";
    double (*p_old_leap_second_table)[LEAPSECOND_COL_NUM] = NULL;
    int line_cnt = 0;
    int scanret = 0;
    
    int utc_y      = 0;
    char   utc_m[24]   = "";
    int utc_d      = 0;
    char   jd_head[24] = "";
    double jd         = 0.0;
    char   leapsecond_head[24] = "";
    double leapsecond  = 0.0;
    
    
    if (NULL != p_leap_second_table)
    {
        /* Leap Second Table has already been initialized. */
        /* return NORMAL; */
        return SUCCESS;
    }
    p_old_leap_second_table = p_leap_second_table;
    
    /* file open */
    p_file = fopen(leap_second_file_name, "r");
    if (NULL == p_file)
    {
        /* error */
        /* return FOPEN_ERROR_LEAPSECOND; */
        return ERROR_FILE_OPEN;
    }
    
    /* file read */
    while (NULL != fgets(buffer, sizeof(buffer), p_file)) 
    {
        if (0 >= strlen(buffer))
        {
            /* A blank line is disregarded. */
            continue;
        }
        
        /* The line is analyzed */
        scanret = sscanf(buffer, "%d %s %d %s %lf %s %lf"
                               , &utc_y, utc_m, &utc_d, jd_head
                               , &jd, leapsecond_head, &leapsecond);

        /* tai93 previously excluded */
        if( LEAPSECOND_TAI93_BEGIN > leapsecond )
        {
            continue;
        }
        
        line_cnt ++;

        DPRINTF("------------------------------------\n");
        DPRINTF("utc : %d %s %d\n", utc_y, utc_m, utc_d);
        DPRINTF("%s : %lf\n", jd_head, jd);
        DPRINTF("%s : %lf\n", leapsecond_head, leapsecond);

        /* memory allocation */
        p_leap_second_table = (double(*)[LEAPSECOND_COL_NUM]) realloc(
            p_old_leap_second_table,
            sizeof(double[LEAPSECOND_COL_NUM]) * line_cnt);
        if (NULL == p_leap_second_table)
        {
            /* error */
            free(p_old_leap_second_table);
            p_old_leap_second_table = NULL;
            fclose(p_file);
            /* return MEM_ERROR_LEAPSECOND; */
            return ERROR_MALLOC;
        }
        p_old_leap_second_table = p_leap_second_table;
        
        p_leap_second_table[line_cnt - 1][TAI93] = ( (jd - JD_TAI93_BEGIN) * SECOND_PER_1DAY
                                                   + (leapsecond - LEAPSECOND_TAI93_BEGIN) );
        p_leap_second_table[line_cnt - 1][MJD] = (jd - 2400000.5);
        p_leap_second_table[line_cnt - 1][LEAPSECOND] = leapsecond;

        DPRINTF("p_leap_second_table[%d][TAI93] = %f\n", line_cnt - 1,
            p_leap_second_table[line_cnt - 1][TAI93]);
        DPRINTF("p_leap_second_table[%d][MJD] = %f\n", line_cnt - 1,
            p_leap_second_table[line_cnt - 1][MJD]);
        DPRINTF("p_leap_second_table[%d][LEAPSECOND] = %.15f\n", line_cnt - 1,
            p_leap_second_table[line_cnt - 1][LEAPSECOND]);
    }
    
    g_leap_second_table_size = line_cnt;
    
    /* file close */
    fclose(p_file);
    
    if (0 >= g_leap_second_table_size)
    {
        /* There was no effective line. */
        /* return NO_EXIST_ERROR_LEAPSECOND_TABLE; */
        return ERROR_NOT_LINE;
    }
    
    /* return NORMAL; */
    return SUCCESS;

}

/** TAI93 TIME -> UTC TIME.
 * It converts it from TAI93 TIME to UTC TIME.
 * 
 * @param tai93_time  [in]TAI93 TIME 
 * @param p_utc_time  [out]UTC TIME 
 * @return NORMAL
 * @return NO_EXIST_ERROR_LEAPSECOND_TABLE
 * @return TIME_ERROR_INPUT_TIME
 * @return TIME_ERROR_LEAPSECOND_TABLE
 */
int com_time_tai93_utc(double tai93_time, utc *p_utc_time)
{
    double utc_second = 0.0;
    double leapsecond_utc_tai = 0.0;
    double leapsecond_utc_tai_tai93begin = 0.0;
    double sec_gmt_to_tai = 0.0;
    
    struct tm *p_temp_utc_time = NULL;
    struct tm tai_begintime = {0};
    time_t utc = 0;
    bool is_just_leapsecond = false;

    
    if (NULL == p_leap_second_table)
    {
        /* error */
        /* return NO_EXIST_ERROR_LEAPSECOND_TABLE; */
        return ERROR_NULL_TABLE;
    }
    
    if (NULL == p_utc_time)
    {
        /* error */
        /* return TIME_ERROR_INPUT_TIME; */
        return ERROR_INPUT_TIME;
    }
    
    /* Get LEAPSECOND(TAI-UTC) */
    if (!get_leapsecond_utc_tai(TAI93, tai93_time,
        &leapsecond_utc_tai, &is_just_leapsecond))
    {
        /* error */
        /* return TIME_ERROR_LEAPSECOND_TABLE; */
        return ERROR_LEAP_SEC;
    }
    
    /* TAI93 standard value (1993/01/01 00:00:00) */
    get_tai93_begintime(&tai_begintime,
        &leapsecond_utc_tai_tai93begin);
    
    /* GMT standard value - TAI93 standard value */
    sec_gmt_to_tai = (double) mktime(&tai_begintime);
    
    /* Get UTC */
    utc_second = tai93_time + sec_gmt_to_tai - leapsecond_utc_tai +
        leapsecond_utc_tai_tai93begin;
    if (is_just_leapsecond)
    {
        /* Just Leap Second (ex: 00:00:00 -> 23:59:59) */
        utc_second -= 1;
    }
    
    utc = (time_t) floor(utc_second);
    
	p_temp_utc_time = localtime(&utc);
    if (NULL == p_temp_utc_time)
    {
        /* error */
        /* return TIME_ERROR_INPUT_TIME; */
        return ERROR_INPUT_TIME;
    }
    
    p_utc_time->year = p_temp_utc_time->tm_year + 1900;
    p_utc_time->month = p_temp_utc_time->tm_mon + 1;
    p_utc_time->day = p_temp_utc_time->tm_mday;
    p_utc_time->hour = p_temp_utc_time->tm_hour;
    p_utc_time->minute = p_temp_utc_time->tm_min;
    
    /* Just Leap Second (ex: 23:59:59 -> 23:59:60) */
    p_utc_time->second = is_just_leapsecond ?
        p_temp_utc_time->tm_sec + 1 : p_temp_utc_time->tm_sec;
    p_utc_time->microsecond = (int) round((utc_second - floor(utc_second))
        * 1000000.);

    return NORMAL;
}


/** UTC TIME -> TAI93 TIME.
 * It converts it from UTC TIME to TAI93 TIME.
 * 
 * @param p_utc_time    [in]UTC TIME 
 * @param p_tai93_time  [out]TAI93 TIME 
 * @return NORMAL
 * @return NO_EXIST_ERROR_LEAPSECOND_TABLE
 * @return TIME_ERROR_INPUT_TIME
 * @return TIME_ERROR_LEAPSECOND_TABLE
 */
int com_time_utc_tai93(const utc *p_utc_time, double *p_tai93_time)
{
    double utc_jd = 0.0;
    double utc_mjd = 0.0;
    double leapsecond_utc_tai = 0.0;
    double leapsecond_utc_tai_tai93begin = 0.0;
    
    struct tm temp_utc_time = {0};
    struct tm tai_begintime = {0};
    bool dummy = false;
    
    
    if (NULL == p_leap_second_table)
    {
        /* error */
        /* return NO_EXIST_ERROR_LEAPSECOND_TABLE; */
        return ERROR_NULL_TABLE;
    }
    
    if ((NULL == p_utc_time) || (NULL == p_tai93_time))
    {
        /* error */
        /* return TIME_ERROR_INPUT_TIME; */
        return ERROR_INPUT_TIME;
    }
    
    /* UTC Y.M.D -> UTC MJD TIME */
    get_jd_and_mjd(p_utc_time->year, p_utc_time->month, 
        p_utc_time->day, &utc_jd, &utc_mjd);
    
    /* Get LEAPSECOND(TAI-UTC) */
    if (!get_leapsecond_utc_tai(MJD, utc_mjd, &leapsecond_utc_tai,
        &dummy))
    {
        /* error */
        /* return TIME_ERROR_LEAPSECOND_TABLE; */
        return ERROR_LEAP_SEC;
    }

    /* UTC STRUCTURE -> TM STRUCTURE */
    temp_utc_time.tm_year = p_utc_time->year - 1900;
    temp_utc_time.tm_mon = p_utc_time->month - 1;
    temp_utc_time.tm_mday = p_utc_time->day;
    temp_utc_time.tm_hour = p_utc_time->hour;
    temp_utc_time.tm_min = p_utc_time->minute;
    temp_utc_time.tm_sec = p_utc_time->second;
    temp_utc_time.tm_isdst = 0;
    
    /* TAI93 standard value (1993/01/01 00:00:00) */
    get_tai93_begintime(&tai_begintime,
        &leapsecond_utc_tai_tai93begin);
    
    /* Get TAI93 */
    *p_tai93_time = difftime(mktime(&temp_utc_time),
        mktime(&tai_begintime))
        + (p_utc_time->microsecond / 1000000.)
        + leapsecond_utc_tai - leapsecond_utc_tai_tai93begin;
    
    
    /* return NORMAL; */
    return SUCCESS;
}


/** Get Julian Day and Modified Julian Day.
 * 
 * @param year  [in] UTC Year
 * @param month [in] UTC month
 * @param day   [in] UTC day
 * @param p_jd  [out]Julian Day
 * @param p_mjd [out]Modified Julian Day
 */
void get_jd_and_mjd(
    int year,
    int month,
    int day,
    double *p_jd,
    double *p_mjd)
{
    int y = 0, m = 0;
    double a = 0.0, b = 0.0, c = 0.0;
    
    
    if (month > 2)
    {
        y = year;
        m = month;
    }
    else
    {
        y = year - 1;
        m = month + 12;
    }

    a = floor((double) y / 100.);
    b = 2 - a + floor(a / 4.);

    if (year > 0)
    {
        c = floor(365.25 * y);
    }
    else
    {
        c = floor(365.25 * y - 0.75);
    }

    *p_jd = 1720994.5 + floor(30.6001 * (m + 1)) + b + c + day;
    *p_mjd = floor(*p_jd - 2400000.5);
    
    return;
}


/** Get TAI93 standard value.
 * (Internal function)
 * @param p_tai_begintime     [out]TAI93 standard value
 * @param p_tai_begin_leapsec [out]Leap Second of TAI93 standard value
 */
static void get_tai93_begintime(
    struct tm *p_tai_begintime,
    double *p_tai_begin_leapsec)
{
    /* TAI93 standard value (1993/01/01 00:00:00) */
    p_tai_begintime->tm_year = 1993 - 1900;
    p_tai_begintime->tm_mon = 1 - 1;
    p_tai_begintime->tm_mday = 1;
    p_tai_begintime->tm_hour = 0;
    p_tai_begintime->tm_min = 0;
    p_tai_begintime->tm_sec = 0;
    p_tai_begintime->tm_isdst = 0;
    
    /* Get Leap Second (UTC - TAI) : 27sec */
    *p_tai_begin_leapsec = LEAPSECOND_TAI93_BEGIN;
    
    return;
}


/** Get LEAPSECOND(TAI-UTC).
 * (Internal function)
 * @param subscript             [in]TAI93 or MJD
 * @param time                  [in]TAI93 TIME or MJD TIME
 * @param p_leapsecond_utc_tai  [out]LEAPSECOND(TAI-UTC)
 * @param p_is_just_leapsecond  [out]true:JUST LEAPSECOND(subscript:TAI93)
 * @return true : exists
 * @return false: no exists
 */
bool get_leapsecond_utc_tai(
    short subscript,
    double time,
    double *p_leapsecond_utc_tai,
    bool *p_is_just_leapsecond)
{
    int i = 0;
    bool is_exists_leapsecond = false;
    
    *p_is_just_leapsecond = false;
    
    if (1 == g_leap_second_table_size)
    {
        /* TABLE SIZE = 1 */
        if (time >= p_leap_second_table[0][subscript])
        {
            *p_leapsecond_utc_tai
                = p_leap_second_table[0][LEAPSECOND];
            is_exists_leapsecond = true;
        }
    }
    else
    {
        /* TABLE SIZE = N */
        for (i = 0; i < (g_leap_second_table_size - 1); i++)
        {
            if ((time >= p_leap_second_table[i][subscript]) &&
                (time < p_leap_second_table[i + 1][subscript]))
            {
                *p_leapsecond_utc_tai
                     = p_leap_second_table[i][LEAPSECOND];
                is_exists_leapsecond = true;
                
                if (TAI93 == subscript)
                {
                    /* TAI93 - 1sec : Just Leap Second
                      (The error margin of double is permitted. )*/
                    if (fabs(floor(time) -
                        floor(p_leap_second_table[i + 1][subscript] - 1.0))
                            < DBL_EPSILON)
                    {
                        /* Just Leap Second */
                        *p_is_just_leapsecond = true;
                    }
                }
                break;
            }
        }

        if ((!is_exists_leapsecond)
             && (time >= p_leap_second_table
                [g_leap_second_table_size - 1][subscript]))
        {
            *p_leapsecond_utc_tai = p_leap_second_table
                [g_leap_second_table_size - 1][LEAPSECOND];
            is_exists_leapsecond = true;
        }
    }
    
    return is_exists_leapsecond;
}


/** Release Leapsecond Table.
 */
void com_release_leapsecond(void)
{
    free(p_leap_second_table);
    p_leap_second_table = NULL;
}

