/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK_hdf.cpp                                              */
/*  Abstract     : I/O toolkit HDF5 access  function                         */
/*  Function List                                                            */
/*      open_file()                 : Open HDF5 file                         */
/*      close_file()                : Close HDF5 file                        */
/*      set_group()                 : Create group                           */
/*      set_dataset()               : Create/Set dataset value               */
/*      set_group_attribute()       : Create/Set group attribute value       */
/*      set_dataset_attribute()     : Create/Set dataset attribute value     */
/*      set_attribute()             : Create/Set attribute value             */
/*      get_dataset()               : Get dataset value                      */
/*      get_group_attribute()       : Get group attribute value              */
/*      get_dataset_attribute()     : Get dataset attribute value            */
/*      get_attribute()             : Get attribute value                    */
/*      get_ds_dims()               : Get dataset dimensions                 */
/*      get_group_attr_dims()       : Get group attribute dimensions         */
/*      get_dataset_attr_dims()     : Get dataset attribute dimensions       */
/*      get_attr_dims()             : Get attribute dimensions               */
/*      get_type_id()               : Get dataset type id                    */
/*      exist_item()                : Check dataset is exists or not         */
/*      exist_group()               : Check group is exists or not           */
/*      exist_item_attributes()     : Check item is exists or not            */
/*      remove_attributes()         : Remove existing attribute              */
/*      set_parent_group()          : Create parent group                    */
/*      create_dataset()            : Create dataset                         */
/*      open_dataset()              : Open dataset                           */
/*      get_attribute_id()          : Get attribute id                       */
/*      get_memtype()               : Get HDF memory type                    */
/*      set_chunk_plist()           : Set property list for chunking         */
/*      set_comp_plist()            : Set property list for compression      */
/*      get_strtype()               : Get memory id of string data type      */
/*      get_type_size()             : Get size of memory id                  */
/*      get_strpad()                : Get string padding                     */
/*      get_float_value_with_slope_and_offset()                              */
/*                                  : Get float value with slope and offset  */
/*      get_double_value_with_slope_and_offset()                             */
/*                                  : Get double value with slope and offset */
/*      get_dataset_attr_float()    : Get float attribute                    */
/*      get_lat_lon_with_interval() : Get latitude/longitude with interval   */
/*      get_data_interval_pixel()   : Get attribute data interval for pixel  */
/*      get_data_interval_line()    : Get attribute data interval for line   */
/*      get_dataset_attr_int()      : Get dataset attribute integer          */
/*      set_float_value_without_slope_and_offset()                           */
/*                              : Set float value without slope and offset   */
/*      set_double_value_without_slope_and_offset()                          */
/*                              : Set double value without slope and offset  */
/*      set_lat_lon_without_interval()                                       */
/*                              : Set latitude/longitude without interval    */
/*      get_group_attr_string_value()                                        */
/*                              : Get group attribute string                 */
/*      get_dataset_attr_string_value()                                      */
/*                              : Get dataset attribute string               */
/*      set_group_attr_string_value()                                        */
/*                              : Set group attribute string                 */
/*      set_dataset_attr_string_value()                                      */
/*                              : Set dataset attribute string               */
/*      get_slope()                 : Get slope value                        */
/*      get_slope_reflectance()     : Get reflectance slope value            */
/*      get_offset()                : Get offset value                       */
/*      get_offset_reflectance()    : Get reclectance offset value           */
/*      get_byteorder()             : Get byte order from exists dataset     */
/*      op_func()                   : Operation function for get_byteorder() */
/*      get_lines_par_scan()        : Get number of lines par scan           */
/*      get_resolution()            : Get resolution                         */
/*      get_highest_resolution()    : Get highest resolution                 */
/*      get_max_line_pixel()                                                 */
/*                          : Get maximum line or pixel for Level-1B product */
/*      get_max_line()              : Get maximum line for Level-1B product  */
/*      get_max_pixel()             : Get maximum pixel for Level-1B product */
/*      get_max_line_pixel_raw_data()                                        */
/*                          : Get maximum line or pixel for Level-1A product */
/*      get_max_line_raw_data()     : Get maximum line for Level-1A product  */
/*      get_max_pixel_raw_data()    : Get maximum pixel for Level-1A product */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
#include <cstdlib>
#include <iostream>
#include "SGTK_hdf.h"
#include "hdf5_hl.h"

using namespace std;

/** open HDF5 file.
 *
 * @param   filename    [string][in]    HDF5 file name.
 * @param   mode        [uint][in]      file open mode.
 * @return HDF5 file id.
 */
hid_t HDF::open_file(const std::string &filename, const uint mode) {

    if( (mode!=SGTK_HDF_ACC_RDONLY) && (mode!=SGTK_HDF_ACC_RDWR) ) {
        return ERROR_PARAM;
    }

    // Turn off error handling
    H5Eset_auto(0, NULL, NULL);

    // Open core file.
    hid_t id_file = H5Fopen(filename.c_str(), mode, H5P_DEFAULT);
    if(0>id_file) {
        // Create core file.
        id_file = H5Fcreate(filename.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
        if(0>id_file) {
              id_file = ERROR_HDF_FILE_OPEN;
        }
        else {
            // byte order flag true set
            byte_order_obj.set_flg(id_file, conf_obj.byte_order);
        }
    }
    else {
        // byte order flag false set
        hid_t id_root = H5Gopen(id_file, "/", H5P_DEFAULT);
        int order = get_byteorder(id_root);
        if(0==order) byte_order_obj.set_flg(id_file, true);
        if(1==order) byte_order_obj.set_flg(id_file, false);
        
    }
    return id_file;
}

/** close HDF5 file.
 *
 * @param id_file   [hid_t][in] HDF5 file id.
 * @return result status.
 */
int HDF::close_file(hid_t id_file) {

    if(0>=id_file) {
        return ERROR_FILE_CLOS;
    }

    // byte order flag data erase
    byte_order_obj.erase_flg(id_file);

    // Close core file.
    H5Fclose(id_file);
    return SUCCESS;
}

/** create group.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   group       [PrdGroupInfo][in]  group info struct.
 * @return  result status.
 */
int HDF::set_group(hid_t id_file, const struct PrdGroupInfo &group) {

    try {
        // Create group.
        int rc = set_parent_group(id_file, group.path);
        if (rc < SUCCESS){
            return rc;
        }

        if(group.attr.size()>0) {
            // Create attribute.
            PrdAttrInfo attr;
            attr.path = group.path;
            attr.attr = group.attr;
            set_group_attribute(id_file, attr);
        }
    }
    catch( const H5::Exception &ex ) {
        return ERROR_SET_GROUP;
    }
    return SUCCESS;
}

/** create/set dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   group       [PrdDatasetInfo][in]    dataset info struct.
 * @return  result status.
 */
int HDF::set_dataset(hid_t id_file, const struct PrdDatasetInfo &dataset) {

    herr_t        status=0;
    hid_t        id_group=0, id_dataset=0, id_dataspace=0, id_memtype=0, id_plist=H5P_DEFAULT;
    std::string    group_path("");
    std::string    ds_path("");
    try {
        group_path = dataset.path.substr(0, dataset.path.find_last_of("/"));
        ds_path = dataset.path.substr(dataset.path.find_last_of("/")+1);
        // Create group.
        if(group_path.empty()) group_path = "/";
        int rc = set_parent_group(id_file, group_path);
        if (rc < SUCCESS){
            return rc;
        }
        id_group = H5Gopen(id_file, group_path.c_str(), H5P_DEFAULT);
        if(H5LTfind_dataset(id_group, ds_path.c_str())==0) {
            id_dataset = create_dataset(id_file, dataset, id_memtype, id_dataspace, id_plist);
        } else {
            id_dataset = open_dataset(id_file, dataset, id_memtype, id_dataspace, id_plist);
        }
        // Write dataset.
        status = H5Dwrite(id_dataset, id_memtype, id_dataspace, H5S_ALL, H5P_DEFAULT, dataset.p_value);
        if(0>status) {
            return ERROR_SET_DATASET;
        }

        // Set dimension label.
        for(unsigned int i=0; i<dataset.label.size(); i++) {
            status = H5DSset_label(id_dataset, i, dataset.label[i].c_str());
            if(0>status) {
                return ERROR_SET_DATASET;
            }
        }

        if(dataset.attr.size()>0) {
            // Create attribute.
            PrdAttrInfo attr;
            attr.path = dataset.path;
            attr.attr = dataset.attr;
            set_dataset_attribute(id_file, attr);
        }
        H5Tclose(id_memtype);
        H5Pclose(id_plist);
        H5Sclose(id_dataspace);
        H5Dclose(id_dataset);
        H5Gclose(id_group);
    }
    catch( const H5::Exception &ex ) {
        return ERROR_SET_DATASET;
    }
    return SUCCESS;
}

/** create/set group attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   attr        [PrdAttrInfo][in]   attribute info struct.
 */
void HDF::set_group_attribute(hid_t id_file, const struct PrdAttrInfo &attr) {
    PrdAttrInfo mod_attr = attr;
    // Remove attribute when it an exists.
    if( true==exist_item_attributes(id_file, attr.path, attr.attr.at(0).name) ) {
        // Set strpad from exists attribute.
        remove_attributes(id_file, attr.path, attr.attr.at(0).name);
    }
    set_attribute(id_file, mod_attr, H5G_GROUP);
}

/** create/set dataset attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   attr        [PrdAttrInfo][in]   attribute info struct.
 */
void HDF::set_dataset_attribute(hid_t id_file, const struct PrdAttrInfo &attr) {
    PrdAttrInfo mod_attr = attr;
    // Remove attribute when it an exists.
    if( true==exist_item_attributes(id_file, attr.path, attr.attr.at(0).name) ) {
        // Set strpad from exists attribute.
        remove_attributes(id_file, attr.path, attr.attr.at(0).name);
    }
    set_attribute(id_file, mod_attr, H5G_DATASET);
}

/** create/set attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   attr        [PrdAttrInfo][in]   attribute info struct.
 * @param   objtype     [H5G_obj_t][in]     type of object.
 */
void HDF::set_attribute(hid_t id_file, const struct PrdAttrInfo &attr, const H5G_obj_t objtype) {

    herr_t    status=0;
    hid_t    id_parent=0, id_dataspace=0, id_memtype=0, id_attribute=0;
    try {
        // Get group or dataset.
        if(H5G_GROUP==objtype) {
            id_parent = H5Gopen(id_file, attr.path.c_str(), H5P_DEFAULT);
        } else if(H5G_DATASET==objtype) {
            id_parent = H5Dopen(id_file, attr.path.c_str(), H5P_DEFAULT);
        } else {
            throw H5::Exception("HDF::set_attribute", "objtype error.");
        }
        // Create attribute.
        vector<PrdAttr>    tmpattr = attr.attr;
        vector<PrdAttr>::iterator it = tmpattr.begin();
        while( it != tmpattr.end() ) {
            id_dataspace = H5Screate_simple(it->rank, (hsize_t*)&it->dims[0], NULL);
            if(H5T_C_S1==it->type) {
                id_memtype = get_strtype(it->length, it->strpad);
            } else {
                id_memtype = get_memtype(it->type);
            }
            id_attribute = get_attribute_id(id_parent, it->name, id_memtype, id_dataspace);

            hid_t id_attr_ds = H5Aget_space(id_attribute);
            hsize_t *maxdims = new hsize_t[it->rank];
            for (int i=0; i<it->rank; i++) maxdims[i] = H5S_UNLIMITED;
            status = H5Sset_extent_simple(id_attr_ds, (hsize_t)it->rank, (hsize_t*)&it->dims[0], maxdims);
            if(0>status) {
                throw H5::Exception("HDF::set_attribute", "H5Sset_extent_simple error.");
            }

            status = H5Awrite(id_attribute, id_memtype, it->p_value);
            if(0>status) {
                throw H5::Exception("HDF::set_attribute", "H5Awrite error.");
            }
            H5Sclose(id_attr_ds);
            H5Sclose(id_dataspace);
            H5Tclose(id_memtype);
            H5Aclose(id_attribute);
            ++it;
			delete maxdims;
        }
        if(H5G_GROUP==objtype) H5Gclose(id_parent);
        if(H5G_DATASET==objtype) H5Dclose(id_parent);
    }
    catch( const H5::Exception &ex ) {
        throw H5::Exception("HDF::set_attribute", "throw error.");
    }
}

/** get dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   p_dataset   [PrdDatasetInfo][out]   dataset info struct.
 */
void HDF::get_dataset(hid_t id_file, struct PrdDatasetInfo *p_dataset) {

    herr_t    status=0;
    hid_t    id_dataset=0, id_dataspace=0, id_memtype=0;
    try {
        // Get dataset.
        id_dataset = H5Dopen(id_file, p_dataset->path.c_str(), H5P_DEFAULT);
        id_dataspace = H5Dget_space(id_dataset);
        if(H5T_C_S1==p_dataset->type) {
            id_memtype = H5Dget_type(id_dataset);
        } else {
            id_memtype = get_memtype(p_dataset->type);
        }

        size_t   nelmts;
        unsigned flags=0, filter_info=1;
        H5Z_filter_t filter_type;
        /* Retrieve filter information. */
        hid_t plist_id = H5Dget_create_plist (id_dataset);
    
        int numfilt = H5Pget_nfilters (plist_id);
     
        for (int i=0; i<numfilt; i++) {
            nelmts = 1;
            unsigned int value=0;
            filter_type = H5Pget_filter2 (plist_id, i, &flags, &nelmts, &value, 0, NULL, &filter_info);
            switch (filter_type) {
                case H5Z_FILTER_DEFLATE:
                    p_dataset->level = value;
                    break;
                default:
                    break;
            }
        }

        // Read dataset.
        status = H5Dread(id_dataset, id_memtype, id_dataspace, H5S_ALL, H5P_DEFAULT, (void*)p_dataset->p_value);
        if(0>status) {
            throw H5::Exception("HDF::get_dataset","ERROR_H5Dread");
        }
        H5Tclose(id_memtype);
        H5Sclose(id_dataspace);
        H5Dclose(id_dataset);
    }
    catch( const H5::Exception &ex ) {
        throw H5::Exception("HDF::get_dataset", "throw error.");
    }
}

/** get group attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   p_attr      [PrdAttrInfo][out]  attribute info struct.
 */
void HDF::get_group_attribute(hid_t id_file, struct PrdAttrInfo *p_attr) {
    get_attribute(id_file, p_attr, H5G_GROUP);
}

/** get dataset attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   p_attr      [PrdAttrInfo][out]  attribute info struct.
 */
void HDF::get_dataset_attribute(hid_t id_file, struct PrdAttrInfo *p_attr) {
    get_attribute(id_file, p_attr, H5G_DATASET);
}

/** get attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   p_attr      [PrdAttrInfo][out]  attribute info struct.
 * @param   objtype     [H5G_obj_t][in]     type of object.
 */
void HDF::get_attribute(hid_t id_file, struct PrdAttrInfo *p_attr, const H5G_obj_t objtype) {

    herr_t    status=0;
    hid_t    id_parent=0, id_memtype=0, id_attribute=0;
    std::vector<H5::Attribute>    attr(1);
    try {
        // Get attribute.
        if(H5G_GROUP==objtype) {
            id_parent = H5Gopen(id_file, p_attr->path.c_str(), H5P_DEFAULT);
        } else if(H5G_DATASET==objtype) {
            id_parent = H5Dopen(id_file, p_attr->path.c_str(), H5P_DEFAULT);
        } else {
            throw H5::Exception("HDF::get_attribute", "objtype error.");
        }
        id_attribute = H5Aopen_name(id_parent, p_attr->attr[0].name.c_str());
        if(H5T_C_S1==p_attr->attr[0].type) {
            id_memtype = H5Aget_type(id_attribute);
            p_attr->attr[0].strpad = H5Tget_strpad(id_memtype);
        } else {
            id_memtype = get_memtype(p_attr->attr[0].type);
        }
        status = H5Aread(id_attribute, id_memtype, (void*)p_attr->attr[0].p_value);
        if(0>status) {
            throw H5::Exception("HDF::get_attribute", "H5Aread error.");
        }
        H5Tclose(id_memtype);
        H5Aclose(id_attribute);
        if(H5G_GROUP==objtype) H5Gclose(id_parent);
        if(H5G_DATASET==objtype) H5Dclose(id_parent);
    }
    catch( const H5::Exception &ex ) {
        throw H5::Exception("HDF::get_attribute", "throw error.");
    }
}

/** get dataset dimensions info.
 * 
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   p_rank      [int*][out]             number of dimension.
 * @param   p_dims      [vector<hsize_t>][out]  number of elemnts, each dimensions.
 * @return result status.
 */
int HDF::get_ds_dims(hid_t id_file, const std::string &path, int *p_rank, vector<hsize_t> *p_dims) {

    hid_t    id_dataset=0, id_dataspace=0;
    hsize_t    size=1;
    try {
        // Get dataspace.
        id_dataset = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
        id_dataspace = H5Dget_space(id_dataset);
        // Get number of dimensions.
        *p_rank = H5Sget_simple_extent_ndims(id_dataspace);
        if(*p_rank<0) {
            return ERROR_HDF_DATASET_GET_DIMS;
        }
        // Get each size of dimensions.
        hsize_t    *dims = new hsize_t[*p_rank];
        int num = H5Sget_simple_extent_dims(id_dataspace, dims, NULL);
        if(num<0) {
            return ERROR_NOT_DIMS;
        }
        // Calculate total size.
        for(int j=0; j<*p_rank; j++) {
            p_dims->push_back(dims[j]);
            size*=dims[j];
        }
        H5Sclose(id_dataspace);
        H5Dclose(id_dataset);
		delete dims;
    }
    catch( const H5::Exception &ex ) {
        return ERROR_NOT_DIMS;
    }
    return size;
}

/** get dimensions of group attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   path        [string][in]        group path.
 * @param   name        [string][in]        attribute name.
 * @param   p_rank      [int*][out]         number of dimension.
 * @param   p_dims      [vector<hsize_t>    number of elements, each dimensions.
 * @return number of elements.
 */
hsize_t HDF::get_group_attr_dims(hid_t id_file, const std::string &path, 
                                                const std::string &name, int *p_rank, vector<hsize_t> *p_dims) {
    return get_attr_dims(id_file, path, H5G_GROUP, name, p_rank, p_dims);
}

/** get dimensions of dataset attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   path        [string][in]        dataset path.
 * @param   name        [string][in]        attribute name.
 * @param   p_rank      [int*][out]         number of dimension.
 * @param   p_dims      [vector<hsize_t>    number of elements, each dimensions.
 * @return number of elements.
 */
hsize_t HDF::get_dataset_attr_dims(hid_t id_file, const std::string &path, 
                                                const std::string &name, int *p_rank, vector<hsize_t> *p_dims) {
    return get_attr_dims(id_file, path, H5G_DATASET, name, p_rank, p_dims);
}

/** get dimensions of attribute.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   path        [string][in]        group/dataset path.
 * @param   objtype     [H5G_obj_t][in]     type of object.
 * @param   name        [string][in]        attribute name.
 * @param   p_rank      [int*][out]         number of dimension.
 * @param   p_dims      [vector<hsize_t>    number of elements, each dimensions.
 * @return number of elements.
 */
hsize_t HDF::get_attr_dims(hid_t id_file, const std::string &path, const H5G_obj_t objtype, 
                                        const std::string &name, int *p_rank, vector<hsize_t> *p_dims) {

    hid_t    id_parent=0, id_dataspace=0, id_attribute=0;
    hsize_t    size=1;
    // Get attribute.
    if(H5G_GROUP==objtype) {
        id_parent = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    } else if(H5G_DATASET==objtype) {
        id_parent = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
    } else {
        throw H5::Exception("HDF::get_attr_dims", "objtype error.");
    }
    id_attribute = H5Aopen(id_parent, name.c_str(), H5P_DEFAULT);
    // Get dataspace.
    id_dataspace = H5Aget_space(id_attribute);
    // Get number of dimensions.
    *p_rank = H5Sget_simple_extent_ndims(id_dataspace);
    if (0 == *p_rank) {
        *p_rank = 1;
    } else if(-1 == *p_rank) {
        throw H5::Exception("HDF::get_attr_dims", "H5Sget_simple_extent_ndims error.");
    }
    // Get each size of dimensions.
    hsize_t    *dims = new hsize_t[*p_rank];
    int num = H5Sget_simple_extent_dims(id_dataspace, dims, NULL);
    if(num<0) {
        throw H5::Exception("HDF::get_attr_dims", "H5Sget_simple_extent_dims error.");
    }
    if (num == 0){
        dims[0] = 1;
    }

    // Calculate total size.
    for(int j=0; j<*p_rank; j++) {
        p_dims->push_back(dims[j]);
        size*=dims[j];
    }
	delete dims;
    H5Sclose(id_dataspace);
    H5Aclose(id_attribute);
    if(H5G_GROUP==objtype) H5Gclose(id_parent);
    if(H5G_DATASET==objtype) H5Dclose(id_parent);
    return size;
}

/** get dataset type id.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    dataset path.
 * @return type id.
 */
hid_t    HDF::get_type_id(hid_t id_file, const std::string &path) {

    hid_t    id_dataset = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
    return H5Dget_type(id_dataset);
}

/** check dataset is exists.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    dataset path.
 * @return result status.
 */
bool HDF::exist_item(hid_t id_file, const std::string &path) {
    bool ret = false;
    hid_t id_parent = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    if(0>id_parent)
    {
        id_parent = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
        if(0>id_parent)
        {
            return false;
        }
        else
        {
            ret = true;
            H5Dclose(id_parent);
        }
    }
    else
    {
        ret = true;
        H5Gclose(id_parent);
    }
    return ret;
}

/** check group is exists.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    group path.
 * @return result status.
 */
bool HDF::exist_group(hid_t id_file, const std::string &path)
{
    bool    ret = true;
    hid_t   id_group = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);

    if (id_group < 0)
    {
        ret = false;
    }
    else
    {
        H5Gclose(id_group);
    }

    return ret;
}

bool HDF::exist_item_attributes(hid_t id_file, const std::string &path, const std::string &name)
{
    bool ret = true;
    htri_t rc = 0;
    hid_t id_parent = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    if(0>id_parent)
    {
        id_parent = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
        if(0>id_parent)
        {
            return false;
        }
        else
        {
            rc = H5Aexists(id_parent, name.c_str());
            if(rc>0) {
                // Exists attribute.
                ret = true;
            } else
            if(0==rc) {
                // Not exist attribute.
                ret = false;
            } else {
                throw H5::Exception("HDF::exist_item_attributes", "H5Aexists error.");
            }
            H5Dclose(id_parent);
        }
    }
    else
    {
        rc = H5Aexists(id_parent, name.c_str());
        if(rc>0) {
            // Exists attribute.
            ret = true;
        } else
        if(0==rc) {
            // Not exist attribute.
            ret = false;
        } else {
            throw H5::Exception("HDF::exist_item_attributes", "H5Aexists error.");
        }
        H5Gclose(id_parent);
    }
    return ret;
}

void HDF::remove_attributes(hid_t id_file, const std::string &path, const std::string &name) {

    try {
        if ( true==exist_item_attributes(id_file, path, name) ) {
            // remove attributes
            herr_t rc = 0;
            hid_t id_object = H5Oopen(id_file, path.c_str(), H5P_DEFAULT);
            rc = H5Adelete(id_object, name.c_str());
            if(0>rc) {
                throw H5::Exception("HDF::remove_attributes", "H5Adelete error.");
            }
            H5Oclose(id_object);
        }
    }
    catch( const H5::Exception &ex ) {
        throw;
    }
}

/** create parent group, if not exists.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][path]  group path.
 * @return result status.
 */
int HDF::set_parent_group(hid_t id_file, const std::string &path) {

    hid_t        id_group=0;
    std::string    parent_path("");
    // Try to open current group.
    id_group = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    if(0>id_group) {
        // Try to create current group.
        id_group = H5Gcreate(id_file, path.c_str(), H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
        if(0>id_group) {
            // Try to create parent group.
            parent_path = path.substr(0, path.find_last_of("/"));
            if(parent_path.size()>0) {
                int rc = set_parent_group(id_file, parent_path);
                if (rc < SUCCESS){
                    return rc;
                }
            }
            // Try to create current group.
            id_group = H5Gcreate(id_file, path.c_str(), H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
            if(0>id_group) {
                return ERROR_SET_PARENT_GROUP;
            }
        }
    }
    H5Gclose(id_group);
    return SUCCESS;
}

/** create dataset.
 * 
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   dataset     [PrdDatasetInfo][in]    dataset info struct.
 * @param   r_memtype   [hid_t][in]             h5 memory type.
 * @param   r_dataspace [hid_t][in]             dataspace id.
 * @param   r_plist     [hid_t][in]             dataset create property list.
 * @return dataset id.
 */
hid_t HDF::create_dataset(hid_t id_file, const PrdDatasetInfo &dataset, hid_t &r_memtype, hid_t &r_dataspace, hid_t &r_plist) {

    hid_t    id_dataset=0;

    // Set compress property.
    set_comp_plist(dataset, r_plist);
    // Create dataspace.
    hsize_t *maxdims = new hsize_t[dataset.rank];
    for (int i=0; i<dataset.rank; i++) maxdims[i] = H5S_UNLIMITED;
    r_dataspace = H5Screate_simple(dataset.rank, &dataset.dims[0], maxdims);
	delete maxdims;
    // Get memory type.
    if(H5T_C_S1==dataset.type) {
        r_memtype = get_strtype(dataset.length, H5T_STR_NULLPAD);
    } else {
        r_memtype = get_memtype(dataset.type);
    }
    // Create dataset.
    id_dataset = H5Dcreate(id_file, dataset.path.c_str(), r_memtype, r_dataspace, H5P_DEFAULT, r_plist, H5P_DEFAULT);
    if(0>id_dataset) {
        throw H5::Exception("HDF::create_dataset", "H5Dcreate error.");
    }

    return id_dataset;
}

/** open dataset.
 * 
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   dataset     [PrdDatasetInfo][in]    dataset info struct.
 * @param   r_memtype   [hid_t][in]             h5 memory type.
 * @param   r_dataspace [hid_t][in]             dataspace id.
 * @param   r_plist     [hid_t][in]             dataset create property list.
 * @return dataset id.
 */
hid_t HDF::open_dataset(hid_t id_file, const PrdDatasetInfo &dataset, hid_t &r_memtype, hid_t &r_dataspace, hid_t &r_plist) {

    herr_t    status=0;
    hid_t    id_dataset=0;

    // Open dataset.
    id_dataset = H5Dopen(id_file, dataset.path.c_str(), H5P_DEFAULT);
    status = H5Dextend(id_dataset, &dataset.dims[0]);
    if(0>status) {
        throw H5::Exception("HDF::open_dataset", "H5Dextend error.");
    }
    status = H5Dset_extent(id_dataset, &dataset.dims[0]);
    if(0>status) {
        throw H5::Exception("HDF::open_dataset", "H5Dset_extent error.");
    }
    // Extent dataspace.
    r_dataspace = H5Dget_space(id_dataset);
    //hsize_t maxdims[dataset.rank];
    hsize_t *maxdims = new hsize_t[dataset.rank];
    for (int i=0; i<dataset.rank; i++) maxdims[i] = H5S_UNLIMITED;
    status = H5Sset_extent_simple(r_dataspace, (hsize_t)dataset.rank, (hsize_t*)&dataset.dims[0], maxdims);
    if(0>status) {
        throw H5::Exception("HDF::open_dataset", "H5Sset_extent_simple error.");
    }
	delete maxdims;
    // Get memory type.
    r_memtype = H5Dget_type(id_dataset);

    return id_dataset;
}

/** get attribute id.
 *
 * @param   id_file     [hid_t][in]         HDF5 file id.
 * @param   name        [string][name]      attribute name.
 * @param   memtype     [hid_t][in]         h5 memory type.
 * @param   dataspace   [hid_t][in]         dataspace id.
 * @return attribute id.
 */
hid_t HDF::get_attribute_id(const hid_t id, const std::string &name, const hid_t memtype, const hid_t dataspace) {

    // Open attribute.
    hid_t    at = H5Aopen(id, name.c_str(), H5P_DEFAULT);
    if(0>at) {
        // Create attribute.
        at = H5Acreate(id, name.c_str(), memtype, dataspace, H5P_DEFAULT, H5P_DEFAULT);
        if(0>at) {
            throw H5::Exception("HDF::get_attribute_id", "H5Acreate error.");
        }
    }
    return at;
}

/** get memory type.
 *
 * @param   type    [hid_t][in]
 * @return h5 memory type.
 */
hid_t HDF::get_memtype(const hid_t type) {

    herr_t    status= 0;
    hid_t    id_memtype=0;

    if(H5T_C_S1==type) {
        // Copy H5T_C_S1 type.
        id_memtype = H5Tcopy(type);
        // Set variable size.
        status = H5Tset_size(id_memtype, 16);
        // Set character set to ASCII.
        status = H5Tset_cset(id_memtype, H5T_CSET_ASCII);
    } else {
        id_memtype = type;
    }
    return id_memtype;
}

/** set property list for chunking.
 *
 * @param   dataset [PrdDatasetInfo][in]    dataset info struct.
 * @param   r_plist [hid_t][out]            chunk property list.
 */
void HDF::set_chunk_plist(const struct PrdDatasetInfo &dataset, hid_t &r_plist) {

    herr_t    status=0;
    hsize_t *cdims = new hsize_t[dataset.rank];
    if (dataset.cdims.empty()) {
        for (int i = 0; i<dataset.rank; i++) {cdims[i] = dataset.dims[i];}
    } else {
        for (int i = 0; i<dataset.rank; i++) {cdims[i] = dataset.cdims[i];}
    }
    // Set chunk size.
    status = H5Pset_chunk(r_plist, dataset.rank, cdims);
	delete cdims;
    if(0>status) {
        throw H5::Exception("HDF::set_chunk_plist", "H5Pset_chunk error.");
    }
}

/** set property list for compression.
 *
 * @param   dataset [PrdDatasetInfo][in]    dataset info struct.
 * @param   r_plist [hid_t][out]            compression property list.
 */
void HDF::set_comp_plist(const struct PrdDatasetInfo &dataset, hid_t &r_plist) {

    herr_t    status=0;
    // Create dataset create property list.
    r_plist = H5Pcreate(H5P_DATASET_CREATE);

    // Set chunk to plist.
    set_chunk_plist(dataset, r_plist);
    status = H5Pset_deflate(r_plist, dataset.level);
    if(0>status) {
        throw H5::Exception("HDF::set_comp_plist", "H5Pset_deflate error.");
    }
}

/** get memory id of string data type.
 *
 * @param   size    [int][in]       string length.
 * @param   strpad  [H5T_str_t][in] type of padding.
 * @return memory id.
 */
hid_t get_strtype(const int size, const H5T_str_t strpad) {
    herr_t  status = 0;
    hid_t   id_memtype = 0;

    // Copy H5T_C_S1 type.
    id_memtype = H5Tcopy(H5T_C_S1);
    if(0>id_memtype) {
        throw H5::Exception("HDF::get_strtype", "H5Tcopy error.");
    }
    // Set variable size.
    status = H5Tset_size(id_memtype, size);
    if(0>status) {
        throw H5::Exception("HDF::get_strtype", "H5Tset_size error.");
    }
	// Set character set to ASCII.
    status = H5Tset_cset(id_memtype, H5T_CSET_ASCII);
    if(0>status) {
        throw H5::Exception("HDF::get_strtype", "H5Tset_cset error.");
    }
    // Set STRPAD.
    status = H5Tset_strpad(id_memtype, strpad);
    if(0>status) {
        throw H5::Exception("HDF::get_strtype", "H5Tset_strpad error.");
    }

    return id_memtype;
}

/** get size of memory id.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][path]  group or dataset path.
 * @param   name        [string][name]  attribute name.
 * @return size of memory id.
 */
hsize_t get_type_size(hid_t id_file, const std::string path, const std::string name) {
    hsize_t size = 0;
    hid_t   id_parent = 0, id_attr = 0, id_memtype = 0;
    H5G_obj_t objtype = H5G_GROUP;

    // Try to open group by path.
    id_parent = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    if(id_parent<0) {
        // Try to open dataset by path.
        id_parent = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
        if(id_parent<0) {
            throw H5::Exception("HDF::get_type_size", "H5Dopen error.");
        }
        objtype = H5G_DATASET;
    }
    id_attr = H5Aopen_name(id_parent, name.c_str());
    id_memtype = H5Aget_type(id_attr);
    size = H5Tget_size(id_memtype);

    H5Tclose(id_memtype);
    H5Aclose(id_attr);
    if(H5G_GROUP==objtype) H5Gclose(id_parent);
    if(H5G_DATASET==objtype) H5Dclose(id_parent);

    return size;
}

/** get string padding.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][path]  group or dataset path.
 * @param   name        [string][name]  attribute name.
 * @return type of padding.
 */
H5T_str_t get_strpad(hid_t id_file, const std::string path, const std::string name) {

    H5T_str_t strpad = H5T_STR_NULLPAD;
    hid_t   id_parent = 0, id_attr = 0, id_memtype = 0;
    H5G_obj_t objtype = H5G_GROUP;

    // Try to open group by path.
    id_parent = H5Gopen(id_file, path.c_str(), H5P_DEFAULT);
    if(id_parent<0) {
        // Try to open dataset by path.
        id_parent = H5Dopen(id_file, path.c_str(), H5P_DEFAULT);
        if(id_parent<0) {
            throw H5::Exception("HDF::get_strpad", "H5Dopen error.");
        }
        objtype = H5G_DATASET;
    }
    id_attr = H5Aopen_name(id_parent, name.c_str());
    id_memtype = H5Aget_type(id_attr);
    strpad = H5Tget_strpad(id_memtype);
    if(0>strpad) {
        throw H5::Exception("HDF::get_strpad", "H5Tget_strpad error.");
    }

    H5Tclose(id_memtype);
    H5Aclose(id_attr);
    if(H5G_GROUP==objtype) H5Gclose(id_parent);
    if(H5G_DATASET==objtype) H5Dclose(id_parent);

    return strpad;
}

/** get float value with slope and offset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<float>][out]    dataset values.
 * @param   type        [int][in]               get value type (radiance or reflectance).
 * @return result status.
 */
int HDF::get_float_value_with_slope_and_offset(hid_t id_file, const std::string &path, std::vector<hsize_t> &dims, std::vector<float> &value, const int type) {

    std::vector<float> temp;
    int rc = get_dataset_numeric_value(id_file, path, dims, temp);
    if(rc < SUCCESS){
        return rc;
    }
    // Get slope and offset value.
    float slope  = 0.0;
    float offset = 0.0;
    if(0==type) {
        slope  = get_slope(id_file, path);
        offset = get_offset(id_file, path);
    } else {
        slope  = get_slope_reflectance(id_file, path);
        offset = get_offset_reflectance(id_file, path);
    }
    if (slope <= 0 || offset < 0) {
        slope = 1;
        offset = 0;
    }
    // Get Mix value
    float vmin = 0;
    bool minret = get_min_attr_value(id_file, path, vmin);
    // Get Max value
    float vmax = 0;
    bool maxret = get_max_attr_value(id_file, path, vmax);
    // Get error value.
    float err = 0;
    bool errret = get_error_value(id_file, path, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    // Get Base
    float vbase = 0;
    bool base_flag = get_base_attr_value(id_file, path, vbase);

    // Get dataset value.
    value.resize(temp.size());
    for(int i=0; i<(int)value.size(); i++) {
        if (range_flag == true){
            if (temp[i] < vmin || vmax < temp[i]){
                value[i] = err;
                continue;
            }
        }
        value[i] = get_mod_value(temp[i], slope, offset);

        if (base_flag == true){
            float out_base = 0;
            set_log_value(value[i], vbase, out_base);
            value[i] = out_base;
        }
    }
    return SUCCESS;
}

/** get double value with slope and offset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<double>][out]   dataset values.
 * @param   type        [int][in]               get value type (radiance or reflectance).
 * @return result status.
 */
int HDF::get_double_value_with_slope_and_offset(hid_t id_file, const std::string &path, std::vector<hsize_t> &dims, std::vector<double> &value, const int type) {

    std::vector<double> temp;
    int rc = get_dataset_numeric_value(id_file, path, dims, temp);
    if(rc < SUCCESS){
        return rc;
    }

    // Get slope and offset value.
    float slope  = 0.0;
    float offset = 0.0;
    if(0==type) {
        slope  = get_slope(id_file, path);
        offset = get_offset(id_file, path);
    } else {
        slope  = get_slope_reflectance(id_file, path);
        offset = get_offset_reflectance(id_file, path);
    }
    if (slope <= 0 || offset < 0) {
        slope = 1;
        offset = 0;
    }

    // Get Mix value
    double vmin = 0;
    bool minret = get_min_attr_value(id_file, path, vmin);
    // Get Max value
    double vmax = 0;
    bool maxret = get_max_attr_value(id_file, path, vmax);
    // Get error value.
    double err = 0;
    bool errret = get_error_value(id_file, path, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    // Get Base
    double vbase = 0;
    bool base_flag = get_base_attr_value(id_file, path, vbase);

    // Get dataset value.
    value.resize(temp.size());
    for(int i=0; i<(int)value.size(); i++) {
        if (range_flag == true){
            if (temp[i] < vmin || vmax < temp[i]){
                value[i] = err;
                continue;
            }
        }
        value[i] = get_mod_value(temp[i], slope, offset);

        if (base_flag == true){
            double out_base = 0;
            set_log_value(value[i], vbase, out_base);
            value[i] = out_base;
        }
    }
    return SUCCESS;
}

/** get float attribute, in dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][out]  number of elements, each dimensions.
 * @param   value       [vector<float>][out]    dataset values.
 * @return result status.
 */
int HDF::get_dataset_attr_float(hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<float> &value) {
    return get_dataset_attr_numeric_value(id_file, path, name, dims, value);
}

/** get latitude/longitude with interval.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   interval    [vector<int>][out]      interval values, each dimensions.
 * @param   dims        [vector<hsize_t>][out]  number of elements, each dimensions.
 * @param   value       [vector<float>][out]    dataset values.
 * @return result status.
 */
int HDF::get_lat_lon_with_interval(hid_t id_file, const std::string &path, std::vector<int> &interval, int type, std::vector<hsize_t> &dims, std::vector<float> &value) {

    // Get interval value.
    if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
        interval.push_back(get_data_interval_line(id_file, path));
    } else
    if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
        interval.push_back(get_resampling_interval(id_file, path));
    } else {
        interval.push_back(1);
    }
    if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
        interval.push_back(get_data_interval_pixel(id_file, path));
    } else
    if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
        interval.push_back(get_resampling_interval(id_file, path));
    } else {
        interval.push_back(1);
    }
    int rc = get_line_pixel_value_with_interval(id_file, path, interval, dims, value, type);
    return rc;
}

/** get data interval for pixel.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    dataset path.
 * @return interval value.
 */
int HDF::get_data_interval_pixel(hid_t id_file, const std::string &path) {
    std::vector<hsize_t> dims;
    std::vector<int> value;
    get_dataset_attr_int(id_file, path, ATTR_DATA_INTERVAL_PIXEL, dims, value);
    return value[0];
}

/** get data interval for line.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    dataset path.
 * @return interval value.
 */
int HDF::get_data_interval_line(hid_t id_file, const std::string &path) {
    std::vector<hsize_t> dims;
    std::vector<int> value;
    get_dataset_attr_int(id_file, path, ATTR_DATA_INTERVAL_LINE, dims, value);
    return value[0];
}

/** get resampling interval.
 *
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][in]    dataset path.
 * @return interval value.
 */
int HDF::get_resampling_interval(hid_t id_file, const std::string &path) {
    std::vector<hsize_t> dims;
    std::vector<int> value;
    get_dataset_attr_int(id_file, path, ATTR_RESAMPLING_INTERVAL, dims, value);
    return value[0];
}

/** get int attribute, in dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<int>][out]      attribute value vector.
 */
void HDF::get_dataset_attr_int(hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<int> &value) {
    get_dataset_attr_numeric_value(id_file, path, name, dims, value);
}

/** set float value without slope and offset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<float>][in]     dataset values.
 * @param   type        [int][in]               get value type (radiance or reflectance).
 * @param   cmplvl      [CompLevel][in]         compression level.
 * @param   chunk       [int][in]               chunk size.
 * @return result status.
 */
int HDF::set_float_value_without_slope_and_offset(
    hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<float> &value, const int type,
        const CompLevel cmplvl, const int chunk) {

    // Get slope and offset value.
    float slope  = 0.0;
    float offset = 0.0;
    if(0==type) {
        slope  = get_slope(id_file, path);
        offset = get_offset(id_file, path);
    } else {
        slope  = get_slope_reflectance(id_file, path);
        offset = get_offset_reflectance(id_file, path);
    }
    if (slope <= 0 || offset < 0) {
        slope = 1;
        offset = 0;
    }

    // Get Mix value
    float vmin = 0;
    bool minret = get_min_attr_value(id_file, path, vmin);
    // Get Max value
    float vmax = 0;
    bool maxret = get_max_attr_value(id_file, path, vmax);
    // Get error value.
    float err = 0;
    bool errret = get_error_value(id_file, path, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    // Get Base
    float vbase = 0;
    bool base_flag = get_base_attr_value(id_file, path, vbase);

    std::vector<float> mod_value;
    for(int i=0; i<(int)value.size(); i++) {
        float out_value = value[i];
        if (base_flag == true){
            set_pow_value(value[i], vbase, out_value);
        }
        mod_value.push_back(get_org_value(out_value, slope, offset));
        if (range_flag == true){
            if (mod_value[i] < vmin || vmax < mod_value[i]){
                mod_value[i] = err;
            }
        }
    }

    std::vector<int> chunks;
    for(int i=0; i<(int)dims.size(); i++) chunks.push_back(chunk);

    std::vector<float> cvt_value(mod_value.size());
    ptr2vec(&mod_value[0], cvt_value, mod_value.size(), byte_order_obj.get_flg(id_file));

    int rc = set_dataset_numeric_value(id_file, path, dims, cvt_value, cmplvl, chunks);
    return rc;
}

/** get float value without slope and offset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<double>][in]    dataset values.
 * @param   type        [int][in]               get value type (radiance or reflectance).
 * @param   cmplvl      [CompLevel][in]         compression level.
 * @param   chunk       [int][in]               chunk size.
 * @return result status.
 */
int HDF::set_double_value_without_slope_and_offset(
    hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<double> &value, const int type,
        const CompLevel cmplvl, const int chunk) {

    // Get slope and offset value.
    float slope  = 0.0;
    float offset = 0.0;
    if(0==type) {
        slope  = get_slope(id_file, path);
        offset = get_offset(id_file, path);
    } else {
        slope  = get_slope_reflectance(id_file, path);
        offset = get_offset_reflectance(id_file, path);
    }
    if (slope <= 0 || offset < 0) {
        slope = 1;
        offset = 0;
    }

    // Get Mix value
    double vmin = 0;
    bool minret = get_min_attr_value(id_file, path, vmin);
    // Get Max value
    double vmax = 0;
    bool maxret = get_max_attr_value(id_file, path, vmax);
    // Get error value.
    double err = 0;
    bool errret = get_error_value(id_file, path, err);

    // flag set
    bool range_flag = true;
    if (minret == true && maxret == true && errret == true) {
        range_flag = true;
    } else {
        range_flag = false;
    }

    // Get Base
    double vbase = 0;
    bool base_flag = get_base_attr_value(id_file, path, vbase);

    std::vector<double> mod_value;
    for(int i=0; i<(int)value.size(); i++) {
        double out_value = value[i];
        if (base_flag == true){
            set_pow_value(value[i], vbase, out_value);
        }
        mod_value.push_back(get_org_value(out_value, slope, offset));
        if (range_flag == true){
            if (mod_value[i] < vmin || vmax < mod_value[i]){
                mod_value[i] = err;
            }
        }
    }

    std::vector<int> chunks;
    for(int i=0; i<(int)dims.size(); i++) chunks.push_back(chunk);


    std::vector<double> cvt_value(mod_value.size());
    ptr2vec(&mod_value[0], cvt_value, mod_value.size(), byte_order_obj.get_flg(id_file));

    int rc = set_dataset_numeric_value(id_file, path, dims, cvt_value, cmplvl, chunks);
    return rc;
}

/** set latitude/longitude without interval.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<float>][in]     dataset values.
 * @param   type        [int][in]               get value type (radiance or reflectance).
 * @return result status.
 */
int HDF::set_lat_lon_without_interval(
    hid_t id_file, const std::string &path, const std::vector<hsize_t> &dims, const std::vector<float> &value, int type) {

    int rc = 0;
    // Set interval value.
    std::vector<int> interval;
    if(2==dims.size()) {
        std::vector<int> interval(2, 1);
        if(dims[0] != 1) {
            if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
                interval.at(0) = get_data_interval_line(id_file, path);
            } else
            if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
                interval.at(0) = (get_resampling_interval(id_file, path));
            } else {
                interval.at(0) = 1;
            }
        }
        if(dims[1] != 1) {
            if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
                interval.at(1) = get_data_interval_pixel(id_file, path);
            } else
            if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
                interval.at(1) = (get_resampling_interval(id_file, path));
            } else {
                interval.push_back(1);
            }
        }
        rc = set_angle_value_without_interval_2d(id_file, path, dims, value, interval, NONE, 0, type);
    } else
    if(3==dims.size()) {
        std::vector<int> interval(3, 1);
        if(dims[1] != 1) {
            if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
                interval.at(1) = get_data_interval_line(id_file, path);
            } else
            if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
                interval.at(1) = (get_resampling_interval(id_file, path));
            } else {
                interval.at(1) = 1;
            }
        }
        if(dims[2] != 1) {
            if(true==exist_item_attributes(id_file, path, ATTR_DATA_INTERVAL_LINE)) {
                interval.at(2) = get_data_interval_pixel(id_file, path);
            } else
            if(true==exist_item_attributes(id_file, path, ATTR_RESAMPLING_INTERVAL)) {
                interval.at(2) = (get_resampling_interval(id_file, path));
            } else {
                interval.at(2) = 1;
            }
        }
        rc = set_angle_value_without_interval_3d(id_file, path, dims, value, interval, NONE, 0, type);
    } else {
        rc =  ERROR_ILLEGAL_DIMS;
    }
    return rc;
}

/** get string attribute, in group.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            group path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<string>][in]    attribute values.
 * @return result status.
 */
int HDF::get_group_attr_string_value(
    hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<std::string> &value) {

    try {
        PrdAttrInfo attrinfo;
        attrinfo.path = path.c_str();
        // Set attribute info.
        PrdAttr attr;
        attr.name = name.c_str();
        attr.type = H5T_C_S1;
        int rc = get_group_attr_dims(id_file, attrinfo.path, attr.name, &attr.rank, &attr.dims);
        if(rc < SUCCESS) {
            return rc;
        }
        // Get elements size.
        size_t size=1;
        if(0!=attr.rank) {
            dims.resize(attr.rank);
            for(int i=0; i<attr.rank; i++) {
                size *= attr.dims[i];
                dims[i] = attr.dims[i];
            }
        } else {
            attr.rank++;
            dims.resize(attr.rank);
            dims[0] = 1;
            size = dims[0];
        }
        char **p_value = new char*[dims[0]*sizeof(char*)];
        attr.p_value = p_value;
        attrinfo.attr.push_back(attr);
        // Get group attribute.
        get_group_attribute(id_file, &attrinfo);
        value.resize(size);
        for(int i=0; i<(int)size; i++) {
            value[i] = p_value[i];
        }
		delete[] p_value;
    }
    catch( const H5::Exception &ex ) {
        return ERROR_GET_GROUP_ATTR;
    }
    return SUCCESS;
}

/** get string attribute, in dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<string>[out]    attribute values.
 * @return result status.
 */
int HDF::get_dataset_attr_string_value(
    hid_t id_file, const std::string &path, const std::string &name, std::vector<hsize_t> &dims, std::vector<std::string> &value) {

    try{
        PrdAttrInfo attrinfo;
        attrinfo.path = path.c_str();
        // Set attribute info.
        PrdAttr attr;
        attr.name = name.c_str();
        attr.type = H5T_C_S1;
        int rc = get_dataset_attr_dims(id_file, attrinfo.path, attr.name, &attr.rank, &attr.dims);
        if(rc < SUCCESS) {
            return rc;
        }
        // Get elements size.
        size_t size=1;
        if(0!=attr.rank) {
            dims.resize(attr.rank);
            for(int i=0; i<attr.rank; i++) {
                size *= attr.dims[i];
                dims[i] = attr.dims[i];
            }
        } else {
            attr.rank++;
            dims.resize(attr.rank);
            dims[0] = 1;
            size = dims[0];
        }
        char **p_value = new char*[dims[0]*sizeof(char*)];
        attr.p_value = p_value;
        attrinfo.attr.push_back(attr);
        // Get dataset attribute.
        get_dataset_attribute(id_file, &attrinfo);
        value.resize(size);
        for(int i=0; i<(int)size; i++) {
            value[i] = p_value[i];
        }
		delete[] p_value;
    }
    catch( const H5::Exception &ex ) {
        return ERROR_GET_DATASET_ATTR;
    }
    return SUCCESS;
}

/** set string attribute, in group.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            group path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<string>][in]    attribute values.
 * @return result status.
 */
int HDF::set_group_attr_string_value(
    hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<std::string> &value) {

    try{
        PrdAttrInfo attrinfo;
        attrinfo.path = path.c_str();
        // Set attribute info.
        PrdAttr attr;
        attr.name = name.c_str();
        attr.type = H5T_C_S1;
        attr.rank = dims.size();
        size_t size=1;
        // Set dimension size.
        for(int i=0; i<attr.rank; i++) {
            attr.dims.push_back(dims[i]);
            size *= dims[i];
        }
        unsigned int length=0;
        for(size_t i=0L; i<size; ++i) {
            if(length<value[i].size()) {
                length = value[i].size();
            }
        }
        char **p_value = new char*[dims[0]*sizeof(char*)];
        for(size_t i=0L; i<size; ++i) {
			p_value[i] = new char[value[i].size() + 1];
            memcpy(p_value[i], value[i].c_str(), value[i].size() + 1);
        }
        attr.length = length;
        attr.p_value = p_value;
        attrinfo.attr.push_back(attr);

        // Set group info.
        PrdGroupInfo groupinfo;
        groupinfo.path = attrinfo.path;
        groupinfo.attr = attrinfo.attr;
        int rc = set_group(id_file, groupinfo);
        if(rc < SUCCESS) {
            return rc;
        }

        // Set group attribute.
        set_group_attribute(id_file, attrinfo);

		delete[] p_value;
    }
    catch( const H5::Exception &ex ) {
        return ERROR_SET_GROUP_ATTR;
    }
    return SUCCESS;
}

/** set string attribute, in dataset.
 *
 * @param   id_file     [hid_t][in]             HDF5 file id.
 * @param   path        [string][in]            dataset path.
 * @param   name        [string][in]            attribute name.
 * @param   dims        [vector<hsize_t>][in]   number of elements, each dimensions.
 * @param   value       [vector<string>][in]    attribute values.
 * @return result status.
 */
int HDF::set_dataset_attr_string_value(
    hid_t id_file, const std::string &path, const std::string &name, const std::vector<hsize_t> &dims, const std::vector<std::string> &value) {

    try{
        PrdAttrInfo attrinfo;
        attrinfo.path = path.c_str();
        // Set attribute info.
        PrdAttr attr;
        attr.name = name.c_str();
        attr.type = H5T_C_S1;
        attr.rank = dims.size();
        size_t size=1L;
        // Set dimension size.
        for(int i=0; i<attr.rank; ++i) {
            attr.dims.push_back(dims[i]);
            size *= dims[i];
        }
        unsigned int length=0;
        for(size_t i=0L; i<size; ++i) {
            if(length<value[i].size()) {
                length = value[i].size();
            }
        }

		//char *p_value = new char[length * size];
		//for(size_t i=0L; i<size; ++i) {
		//    memcpy(&p_value[i*length], value[i].c_str(), value[i].size() + 1);
		//}
		char **p_value = new char*[dims[0] * sizeof(char*)];
		for (size_t i = 0L; i<size; ++i) {
			p_value[i] = new char[value[i].size() + 1];
			memcpy(p_value[i], value[i].c_str(), value[i].size() + 1);
		}
        attr.length = length;
        attr.p_value = p_value;
        attrinfo.attr.push_back(attr);

        // Set dataset attribute.
        set_dataset_attribute(id_file, attrinfo);

		delete[] p_value;
    }
    catch( const H5::Exception &ex ) {
        return ERROR_SET_DATASET_ATTR;
    }
    return SUCCESS;
}

/** get slope value from dataset attribute.
 * 
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][out]   dataset path.
 * @return slope value.
 */
float HDF::get_slope( hid_t id_file, const std::string &path) {

    std::vector<hsize_t> dims;
    std::vector<float>   value;

    int     elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.correct_slope_attr_name);
    if (elment != -1)
    {
        int rc = get_dataset_attr_float(id_file, path, conf_obj.correct_slope_attr_name.at(elment), dims, value);
        if (rc < SUCCESS){
            return (float)ERROR_GET_SLOPE;
        }
    }
    else
    {
        return ERROR_NOT_FIND_SLOPE;
    }
    return value[0];
}

/** get slope reflectance value from dataset attribute.
 * 
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][out]   dataset path.
 * @return slope reflectance value.
 */
float HDF::get_slope_reflectance(hid_t id_file, const std::string &path) {

    std::vector<hsize_t> dims;
    std::vector<float>   value;

    int     elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.correct_slope_ref_attr_name);
    if (elment != -1)
    {
        int rc = get_dataset_attr_float(id_file, path, conf_obj.correct_slope_ref_attr_name.at(elment), dims, value);
        if (rc < SUCCESS){
            return (float)ERROR_GET_SLOPE_REF;
        }
    }
    else
    {
        return ERROR_NOT_FIND_SLOPE_REF;
    }
    return value[0];
}

/** get offset value from dataset attribute.
 * 
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][out]   dataset path.
 * @return offset value.
 */
float HDF::get_offset(hid_t id_file, const std::string &path) {

    std::vector<hsize_t> dims;
    std::vector<float>   value;

    int     elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.correct_offset_attr_name);
    if (elment != -1)
    {
        int rc = get_dataset_attr_float(id_file, path, conf_obj.correct_offset_attr_name.at(elment), dims, value);
        if (rc < SUCCESS){
            return (float)ERROR_GET_OFFSET;
        }
    }
    else
    {
        return ERROR_NOT_FIND_OFFSET;
    }
    return value[0];
}

/** get offset reflectance value from dataset attribute.
 * 
 * @param   id_file     [hid_t][in]     HDF5 file id.
 * @param   path        [string][out]   dataset path.
 * @return offset reflectance value.
 */
float HDF::get_offset_reflectance(hid_t id_file, const std::string &path) {

    std::vector<hsize_t> dims;
    std::vector<float>   value;

    int     elment = find_attr_name_data(id_file, (char *)path.c_str(), conf_obj.correct_offset_ref_attr_name);
    if (elment != -1)
    {
        int rc = get_dataset_attr_float(id_file, path, conf_obj.correct_offset_ref_attr_name.at(elment), dims, value);
        if (rc < SUCCESS){
            return (float)ERROR_GET_OFFSET_REF;
        }
    }
    else
    {
        return ERROR_NOT_FIND_OFFSET_REF;
    }
    return value[0];
}

/** get byte order at exists hdf file.
 * 
 * @param   loc_id  [hid_t][in] hdf location id.
 * @return byte order flag (little endian:true / big endian:false).
 */
int get_byteorder(hid_t loc_id) {
    int isLittle = 2;
    herr_t status = H5Ovisit(loc_id, H5_INDEX_NAME, H5_ITER_NATIVE, op_func, (void*)&isLittle);
    if(status<0) {
        return status;
    }
    return isLittle;
}

/** operation function for get_byteorder.
 *
 * @param   loc_id          [hid_t][in]         HDF5 location id.
 * @param   name            [char*][in]         location name.
 * @param   info            [H5O_info_t][in]    H5 object info.
 * @param   operator_data   [void*][in/out]     user parameter pointer.
 * @return result status.
 */
herr_t op_func(hid_t loc_id, const char *name, const H5O_info_t *info, void *operator_data) {
    if(name[0] == '.') {
        ;
    } else {
        hid_t type;
        hid_t group;
        H5T_order_t order;
        std::string path;
        switch(info->type) {
        case H5O_TYPE_GROUP:
            group = H5Gopen(loc_id, path.c_str(), H5P_DEFAULT);
            *((int*)operator_data) = get_byteorder(group);
            break;
        case H5O_TYPE_DATASET:
            path.assign(name);
            type = HDF::get_type_id(loc_id, path);
            order = H5Tget_order(type);
            switch(order) {
            case H5T_ORDER_LE:
                *((int*)operator_data) = 0;
                return 1;
                break;
            case H5T_ORDER_BE:
                *((int*)operator_data) = 1;
                return 1;
                break;
            default:
                return FAIL;
                break;
            }
            break;
        case H5O_TYPE_NAMED_DATATYPE:
            break;
        default:
            break;
        }
    }
    return SUCCESS;
}

/** get number of lines par scan.
 *
 * @param   granule_id  [char*][in] granule-id.
 * @return number of lines.
 */
int get_lines_par_scan(char *granule_id, const std::string &path) {

    int line = 0;

    if( (granule_id[26]=='1') && (granule_id[27]=='A') ) {
        // level 1A product.
        if( (granule_id[31]=='V') && (granule_id[32]=='N') && (granule_id[33]=='R') ) {
            // VNR sensor.
            line = 1;
        } else if( (granule_id[31]=='P') && (granule_id[32]=='O') && (granule_id[33]=='L') ) {
            // VNR sensor.
            line = 1;
        } else if( (granule_id[31]=='I') && (granule_id[32]=='R') && (granule_id[33]=='S') ) {
            bool is_swi = false;
            if( (std::string::npos!=path.find("SW1",0)) || 
                (std::string::npos!=path.find("SW2",0)) || 
                (std::string::npos!=path.find("SW4",0)) ||
                (std::string::npos!=path.find("SW3",0)) ) {
                is_swi = true;
            } else
            if( (std::string::npos!=path.find("TI1",0)) ||
                (std::string::npos!=path.find("TI2",0)) ) {
                is_swi = false;
            } else {
                return ERROR_PARAM;
            }

            // IRS sensor.
            switch(granule_id[35]) {
            case 'Q':
                // 250m.
                line = 20;
                break;
            case 'M':
                if(true==is_swi) {
                    // SWI
                    // 250m.
                    line = 20;
                } else {
                    // TIR
                    // 500m.
                    line = 10;
                }
                break;
            case 'X':
                if(true==is_swi) {
                    // SWI
                    // 250m.
                    line = 20;
                } else {
                    // TIR
                    // 1000m.
                    line = 5;
                }
                break;
            case 'Y':
                if(true==is_swi) {
                    // SWI
                    // 1000m.
                    line = 5;
                } else {
                    // TIR
                    // 250m.
                    line = 20;
                }
                break;
            case 'H':
                if(true==is_swi) {
                    // SWI
                    // 1000m.
                    line = 5;
                } else {
                    // TIR
                    // 500m.
                    line = 10;
                }
                break;
            case 'K':
            case 'L':
                // 1000m.
                line = 5;
                break;
            default:
                return ERROR_PARAM;
                break;
            }
        } else {
            return ERROR_PARAM;
        }
    } else
    if( (granule_id[26]=='1') && (granule_id[27]=='B') ) {
        // level 1B product.
        if( (granule_id[31]=='V') && (granule_id[32]=='N') && (granule_id[33]=='R') ) {
            // VNR sensor.
            line = 1;
        } else if( (granule_id[31]=='P') && (granule_id[32]=='O') && (granule_id[33]=='L') ) {
            // VNR sensor.
            line = 1;
        } else if( (granule_id[31]=='I') && (granule_id[32]=='R') && (granule_id[33]=='S') ) {
            // IRS sensor.
            switch(granule_id[35]) {
            case 'Q':
            case 'M':
            case 'X':
            case 'Y':
                // 250m.
                line = 20;
                break;
            case 'H':
                // 500m.
                line = 10;
                break;
            case 'K':
            case 'L':
                // 1000m.
                line = 5;
                break;
            default:
                return ERROR_PARAM;
                break;
            }
        } else {
            return ERROR_PARAM;
        }

    } else if( (granule_id[26]=='L') && (granule_id[27]=='2') ) {
        // level 2 or level 3 product.
        line = 1;
    } else if( (granule_id[26]=='3') && (granule_id[27]=='B') ) {
        // level 2 or level 3 product.
        line = 1;
    } else if( (granule_id[26]=='3') && (granule_id[27]=='M') ) {
        // level 2 or level 3 product.
        line = 1;
    } else {
        return ERROR_PARAM;
    }

    return line;
}


/** get resolution.
 *
 * @param   path        [string][in]    dataset path.
 * @return resolution size (meter).
 * @note If the attribute does not exist, 0 is returned.
 */
int get_resolution(hid_t id_file, const std::string &path) {

    int result = 0;
    std::vector<unsigned char> resolution;
    std::vector<hsize_t> dims;

    if( (std::string::npos!=path.find("SW1",0)) || 
        (std::string::npos!=path.find("SW2",0)) || 
        (std::string::npos!=path.find("SW4",0)) ||
        (std::string::npos!=path.find("Lt_SW01",0)) || 
        (std::string::npos!=path.find("Lt_SW02",0)) || 
        (std::string::npos!=path.find("Lt_SW04",0)) ) {
        result = 1000;
    } else if( (std::string::npos!=path.find("SW3",0)) ||
               (std::string::npos!=path.find("Lt_SW03",0)) ) {

        int rc =  HDF::get_dataset_numeric_value(id_file, "/Ancillary_data/SWI_DSP_FPGA/Resolution_status", dims, resolution);
        if(rc < SUCCESS){
            return rc;
        }
        switch(resolution.at(0)) {
        case 1:
            result = 250;
            break;
        case 3:
            result = 1000;
            break;
        default:
            break;
        }

    } else if( (std::string::npos!=path.find("TI1",0)) ||
               (std::string::npos!=path.find("TI2",0)) ||
               (std::string::npos!=path.find("Lt_TI01",0)) ||
               (std::string::npos!=path.find("Lt_TI02",0)) ) {

        int rc = HDF::get_dataset_numeric_value(id_file, "/Ancillary_data/TIR_DSP_FPGA/Resolution_status", dims, resolution);
        if(rc < SUCCESS){
            return rc;
        }
        switch(resolution.at(0)) {
        case 1:
            result = 250;
            break;
        case 2:
            result = 500;
            break;
        case 3:
            result = 1000;
            break;
        default:
            break;
        }

    }

    return result;
}

/** get highest resolution.
 * Returns the highest resolution in the product file.
 * @return highest resolution size in product (meter).
 */
int get_highest_resolution(hid_t id_file) {

    std::vector<hsize_t> dims;
    std::vector<unsigned char> swi;
    std::vector<unsigned char> tir;
    int resolution_swi = 0;
    int resolution_tir = 0;
    std::string sw3("SW3");
    std::string ti1("TI1");

    resolution_swi = get_resolution(id_file, sw3);
    resolution_tir = get_resolution(id_file, ti1);

    int result = 0;
    if( resolution_swi <= resolution_tir ) {
        result = resolution_swi;
    } else {
        result = resolution_tir;
    }

    return result;
}

/** get maximum line or pixel for Level-1B product.
 *
 * @param   path        [string][in]    dataset path.
 * @param   name        [string][in]    attribute name.
 * @return maximum line or pixel in the product file.
 */
int get_max_line_pixel(hid_t id_file, const std::string &path, const std::string &name) {

    int result = 0;

    std::vector<hsize_t> dims;
    std::vector<int> value;
    int rc = HDF::get_group_attr_numeric_value(id_file, "/Image_data", name, dims, value);
    if(rc < SUCCESS){
        return 0;
    }

    if( (std::string::npos!=path.find("Lt_SW01",0)) || 
        (std::string::npos!=path.find("Lt_SW02",0)) || 
        (std::string::npos!=path.find("Lt_SW04",0)) ||
        (std::string::npos!=path.find("Lt_SW03",0)) ||
        (std::string::npos!=path.find("Lt_TI01",0)) ||
        (std::string::npos!=path.find("Lt_TI02",0)) ) {
        int reso = get_resolution(id_file, path);
        int highest_reso = get_highest_resolution(id_file);
        if( (highest_reso==0) || (reso==0) ) {
            result = value.at(0);
        } else {
            result = value.at(0) * ((double)highest_reso/(double)reso);
        }

    } else {
        result = value.at(0);
    }

    return result;
}

/** get maximum line.
 * Get maximum line from "/Image_data/Number_of_lines" attribute.
 * @param   path        [string][in]    dataset path.
 * @retrun maximum line in the product file.
 * @note If the attribute does not exist, 0 is returned.
 */
int get_max_line(hid_t id_file, const std::string &path) {
    return get_max_line_pixel(id_file, path, "Number_of_lines");
}

/** get maximum pixel.
 * Get maximum pixel from "/Image_data/Number_of_pixels" attribute.
 * @param   path        [string][in]    dataset path.
 * @retrun maximum pixel in the product file.
 * @note If the attribute does not exist, 0 is returned.
 */
int get_max_pixel(hid_t id_file, const std::string &path) {
    return get_max_line_pixel(id_file, path, "Number_of_pixels");
}

/** get maximum line or pixel for Level-1A product.
 *
 * @param   path        [string][in]    dataset path.
 * @param   name        [string][in]    attribute name.
 * @return maximum line or pixel in the product file.
 */
int get_max_line_pixel_raw_data(hid_t id_file, const std::string &path, const std::string &name) {

    int result = 0;

    std::vector<hsize_t> dims;
    std::vector<int> value;
    int rc = HDF::get_group_attr_numeric_value(id_file, "/Raw_data", name, dims, value);
    if(rc < SUCCESS){
        return 0;
    }

    if( (std::string::npos!=path.find("SW1",0)) || 
        (std::string::npos!=path.find("SW2",0)) || 
        (std::string::npos!=path.find("SW4",0)) ||
        (std::string::npos!=path.find("SW3",0)) ||
        (std::string::npos!=path.find("TI1",0)) ||
        (std::string::npos!=path.find("TI2",0)) ) {
        int reso = get_resolution(id_file, path);
        int highest_reso = get_highest_resolution(id_file);
        if( (highest_reso==0) || (reso==0) ) {
            result = value.at(0);
        } else {
            result = value.at(0) * ((double)highest_reso/(double)reso);
        }

    } else {
        result = value.at(0);
    }

    return result;
}

/** get maximum line.
 * Get maximum line from "/Raw_data/Number_of_lines" attribute.
 * @param   path        [string][in]    dataset path.
 * @retrun maximum line in the product file.
 * @note If the attribute does not exist, 0 is returned.
 */
int get_max_line_raw_data(hid_t id_file, const std::string &path) {
    return get_max_line_pixel_raw_data(id_file, path, "Number_of_lines");
}

/** get maximum pixel.
 * Get maximum pixel from "/Raw_data/Number_of_pixels" attribute.
 * @param   path        [string][in]    dataset path.
 * @retrun maximum pixel in the product file.
 * @note If the attribute does not exist, 0 is returned.
 */
int get_max_pixel_raw_data(hid_t id_file, const std::string &path) {
    return get_max_line_pixel_raw_data(id_file, path, "Number_of_pixels");
}
