/*****************************************************************************/
/*  Copyright (C) NEC Corporation 2015 All rights reserved.                  */
/*  File Name    : SGTK_latlon.cpp                                           */
/*  Abstract     : I/O toolkit calculate latitude/longitude binary  function */
/*  Function List                                                            */
/*      calcLatLon_eqa_tile()        : Calculate EQA tile binary             */
/*      calcLatLon_eqa_binning()     : Calculate EQA binning binary          */
/*      calcLatLon_eqr()             : Calculate EQR binary                  */
/*      calcLatLon_ps()              : Calculate PS binary                   */
/*                                                                           */
/*  Author     NEC                                                           */
/*  Date       2015/03/20                                                    */
/*  Version    1.00                                                          */
/*  Notice                                                                   */
/*****************************************************************************/
#include "SGTK_latlon.h"

/** get latitude and longitude matrix (EQA tile).
 *
 * @param   resolution  [int][in]       resolution.
 * @param   h           [int][in]       tile number of horizontal(00 - 35).
 * @param   v           [int][in]       tile number of virtical(00 - 17).
 * @param   latlon      [float*][out]   binary matrix of latitude and longitude.
 */
void calcLatLon_eqa_tile(int resolution, int h, int v, float *latlon){
    
    float *output = latlon;
    int N;  // タイルの縦横サイズ                   

    if( resolution == 0 )       N = 1200;   // 1km
    else if( resolution == 1  ) N = 2400;   // 500m
    else                        N = 4800;   // 250m

    FLOAT lat, lon;
    for(float x=0.5; x<=N+0.5; x++){
        for(float y=0.5; y<=N+0.5; y++ ){
    
            lat = 90 - 10 * v - 10*(y-0.5)/N;
            if( (lat < -90) || (90 < lat) ) {
                lat = -999;
            }
                    
            //lon = (10 * (x - 0.5) / N + 10 * h - 180) / cos(lat*M_PI / 180);
            lon = (10 * (x - 0.5) / N + 10 * h - 180) / cos(lat*3.14159265358979323846 / 180);
            if( (lon < -180) || (180 < lon) ) {
                lon = -999;
            }

            *(output) = lat;
            *(output+1) = lon;
            output+=2;
        }
    }
}
/** get latitude and longitude matrix (EQA binning).
 *
 * @param   resolution  [int][in]       resolution.
 * @param   latlon      [float*][out]   binary matrix of latitude and longitude.
 */
void calcLatLon_eqa_binning(int resolution, float *latlon){
    
    float *output = latlon;
    double d = resolution == 0 ? 1.0/12 : 1.0/24;
    int NL = (int)round(180 / d);
    int NP0 = 2 * (int)round(180 / d);
    int NPi;
    FLOAT lat, lon; // latitude & longitude of bin[i]
    int col_min, col_max, col_0;
    

    for(int lin = NL; lin >= 1; lin--){

        //NPi = (int)round(NP0 * sin((lin - 0.5) * d * M_PI / 180));
        NPi = (int)round(NP0 * sin((lin - 0.5) * d * 3.14159265358979323846 / 180));

        col_0 = (int)( round(NP0/2.0 + 1) );
        col_min = (int)( col_0 - ceil(NPi/2.0) );
        col_max = (int)( col_0 + floor(NPi/2.0) );
        
        // 中心から東
        for( int col = col_0; col < col_max; col++ ){

            lat = 90 - (lin - 0.5) * d;
            if( (lat < -90) || (90 < lat) ) {
                lat = -999;
            }
            
            lon = 360 / NPi * ( col - NP0/2 - 0.5 );                
            if( (lon < -180) || (180 < lon) ) {
                lon = -999;
            }

            *(output) = lat;
            *(output+1) = lon;
            output+=2;
    
        }

        // 西から中心
        for( int col = col_min; col < col_0; col++ ){

            lat = 90 - (lin - 0.5) * d;
            if( (lat < -90) || (90 < lat) ) {
                lat = -999;
            }
            
            lon = 360 / NPi * ( col - NP0/2 - 0.5 );                
            if( (lon < -180) || (180 < lon) ) {
                lon = -999;
            }

            *(output) = lat;
            *(output+1) = lon;
            output+=2;
    
        }

    }

}


/** get latitude and longitude matrix (EQR).
 *
 * @param   resolution  [int][in]       resolution.
 * @param   latlon      [float*][out]   binary matrix of latitude and longitude.
 */
void calcLatLon_eqr(int resolution, float *latlon){
    
    float *output = latlon;
    int d = resolution == 0 ? 12 : 24;

    int ROW = 180 * d;  //[pix]
    int COL = 360 * d;  //[pix]

    FLOAT delta = 1.0 / d;  //[deg]

    FLOAT lat, lon;

    for(int i=0; i<ROW; i++){
        for(int j=0; j<COL; j++){

            lat = 90 - delta * i - delta/2;
            if( (lat < -90) || (90 < lat) ) {
                lat = -999;
            }
            lon = -180 + delta * j + delta/2;
            if( (lon < -180) || (180 < lon) ) {
                lon = -999;
            }
        
            *(output) = lat;
            *(output+1) = lon;
            output+=2;

        }
    }

}


/** get latitude and longitude matrix (PS).
 *
 * @param   resolution  [int][in]       resolution.
 * @param   latlon      [float*][out]   binary matrix of latitude and longitude.
 */
void calcLatLon_ps(int resolution, float *latlon){
    
    float *output = latlon;

    double e = 1.0/298.257223563;   // 地球の扁平率(WGS84)

    int N = 5500;
    bool isNorth = true;

    switch(resolution){
        case 0: // North    1/12
            N /= 2;
            isNorth = true;
            break;
        case 1: // North    1/24
            isNorth = true;
            break;  
        case 2: // South    1/12
            N /= 2;
            isNorth = false;
            break;
        case 3: // South    1/24
            isNorth = false;
            break;
    }

    double R = N / 2.0;

    FLOAT lat = 0.0, lon = 0.0, x = 0.0, y = 0.0;

    for(int i=1; i<=N; i++){
        for(int j=1; j<=N; j++){

            x = (j-0.5) - N/2;
            y = (N-i+0.5) - N/2;

            if( sqrt(x*x + y*y) > R  ){     // 地表がない = 欠損値

                lat = -999;
                lon = -999;

            }else if( x==0 && y==0 ){       // 場合分け
                
                if(isNorth) lat = 90;
                else        lon = -90;

                lon = 0;

            }else{

                lat = atan2(1, sqrt( (1-e*e) * (2*R/sqrt(x*x+y*y)-1) ) );
                if(!isNorth) lat *= -1;             

                lon = atan2(-x, y);
                if(!isNorth) lon *= -1;

            }

            *(output) = lat;
            *(output+1) = lon;
            output+=2;

        }
    }
}
